@@ -24,10 +24,10 @@ source directory.
 LILO_EXTRA extra;
 char *identify = NULL;
 int boot_dev_nr, raid_index, do_md_install;
-int verbose = 0,test = 0,compact = 0,linear = 0, raid_flags = 0, zflag = 0,
-      ireloc = 0, force_fs = 0, extended_pt = 0, query = 0,
+int verbose = 0, test = 0, compact = 0, linear = 0, raid_flags = 0, zflag = 0,
+      ireloc = 0, force_fs = 0, force_raid = 0, extended_pt = 0, query = 0,
       nowarn = 0, lba32 = 0, autoauto = 0, passw = 0, geometric = 0, eflag = 0;
-int colormax = 15;
+int colormax = 15, warnings = 0;
 DL_BIOS bios_passes_dl = DL_NOT_SET;
 
 #if !__MSDOS__
@@ -65,6 +65,24 @@ int curr_prt_map;
 }
 
 
+/*volatile*/ void warn(char *fmt,...)
+{
+    va_list ap;
+
+    warnings++;
+    if (nowarn > 0) return;
+    
+    fflush(stdout);
+    fprintf(errstd,"Warning: ");
+    va_start(ap,fmt);
+    vfprintf(errstd,fmt,ap);
+    va_end(ap);
+    fputc('\n',errstd);
+    
+    return;
+}
+
+
 void *alloc(int size)
 {
     void *this;
