@@ -1,7 +1,7 @@
 /* config.h  -  Configurable parameters */
 /*
 Copyright 1992-1998 Werner Almesberger.
-Copyright 1999-2005 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
@@ -13,6 +13,13 @@ source directory.
 #ifndef CONFIG_H
 #define CONFIG_H
 
+/* undefine LCF_DEVMAPPER if the library is not present */
+#ifdef LCF_DEVMAPPER
+# ifndef HAS_LIBDEVMAPPER_H
+#  undef LCF_DEVMAPPER
+# endif
+#endif
+
 #if !__MSDOS__
 #if !defined(__GLIBC__) || (__GLIBC__ < 2) || \
 	 !defined(__GLIBC_MINOR__) || (__GLIBC_MINOR__ < 1)
@@ -46,7 +53,9 @@ source directory.
 #define DFL_KEYTAB  BOOT_DIR "/us.ktl"	/* default keytable */
 #endif
 
-#define DEV_DIR	    "/dev" /* devices directory */
+#define DEV_DIR	    	"/dev" 	/* devices directory (/dev/mem &c.) */
+#define DEV_DISK_DIR  	DEV_DIR	/* disk devices are here */
+                                /* alternate might be "/dev/disk" */
 
 #else /* MSDOS */
 #define CFG_DIR	    "C:\\ETC"		/* location of configuration files */
