@@ -1,15 +1,13 @@
 /* device.c  -  Device access */
 /*
 Copyright 1992-1997 Werner Almesberger.
-Copyright 1999-2005 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
 source directory.
 
 */
-#define DEBUG 0
-#define ATAKEYWORD 0
 
 #define _GNU_SOURCE
 #include <unistd.h>
@@ -34,7 +32,8 @@ source directory.
 #include "partition.h"
 #include "cfg.h"
 #include "probe.h"
-#include "ata-u.h"
+#include "md-int.h"
+
 
 typedef struct _cache_entry {
     const char *name;
@@ -89,6 +88,7 @@ static int scan_dir(ST_BUF *next,DEVICE *dev,char *parent,int number)
     ST_BUF st,*walk;
     char *start;
 
+    if (verbose >= 5) printf("scan_dir: %s\n", parent);
     st.next = next;
     if ((dp = opendir(parent)) == NULL)
 	die("opendir %s: %s",parent,strerror(errno));
@@ -101,8 +101,13 @@ static int scan_dir(ST_BUF *next,DEVICE *dev,char *parent,int number)
 		(void) closedir(dp);
 		return 1;
 	    }
+#if 0
 	    if (S_ISDIR(dev->st.st_mode) && strcmp(dir->d_name,".") &&
 	      strcmp(dir->d_name,"..")) {
+#else
+    /* stay out of all hidden directories (good for 2.6 kernel) */
+	    if (S_ISDIR(dev->st.st_mode) && (dir->d_name)[0] != '.') {
+#endif
 		for (walk = next; walk; walk = walk->next)
 		    if (stat_equal(&walk->st,&st.st)) break;
 		if (!walk && scan_dir(&st,dev,parent,number)) {
@@ -156,7 +161,7 @@ static int yesno = 0;
     struct stat st;
    
     if (!yesno) {
-	if (stat("/dev/.devfsd", &st) < 0)  yesno=1;
+	if (stat(DEV_DIR "/.devfsd", &st) < 0)  yesno=1;
 	else yesno=2;
     }
    
@@ -188,7 +193,7 @@ static void cache_add(const char *name,int number)
 
 int dev_open(DEVICE *dev,int number,int flags)
 {
-    char name[PATH_MAX+1];
+    char name[PATH_MAX];
     ST_BUF st;
     int count;
 
@@ -223,7 +228,7 @@ int dev_open(DEVICE *dev,int number,int flags)
 	else cache_add(name,number);
     }
 #if BETA_TEST
-	if (verbose >= 3) printf("stat-ing %s\n", name);
+	if (verbose >= 4) printf("stat-ing %s\n", name);
 	fflush(stdout);
 	stat(name, &st.st);
 #endif
@@ -253,8 +258,9 @@ void dev_close(DEVICE *dev)
 
 void cache_scsi (char *name, int major)
 {
-    char tem[80], format[80];
+    char tem[PATH_MAX], format[PATH_MAX];
     int i, j, dev;
+    int k = strlen(DEV_DISK_DIR) + 3;
     
     for (i=15; i>=0; i--) {
 	dev = MKDEV(major,(i<<4));
@@ -269,7 +275,7 @@ void cache_scsi (char *name, int major)
 #endif
 	{
 	    strcpy(format, name);
-	    format[7] += i;
+	    format[k] += i;
 	    cache_add(format, dev);
 	    strcat(format, "%d");
 	}
@@ -283,19 +289,20 @@ void cache_scsi (char *name, int major)
 
 void cache_ide (char *name, int major)
 {
-    char tem[80], tem2[80];
-    char format[80];
+    char tem[PATH_MAX], tem2[PATH_MAX];
+    char format[PATH_MAX];
     char *disc;
     int i, dev, minor, host, bus, target;
+    int j = strlen(DEV_DISK_DIR) + 3;
     
-    i = name[7] - 'a';
+    i = name[j] - 'a';
     target = i&1;
     minor = target<<6;
     i >>= 1;
     bus = i&1;
     host = i&(-2);
     if (is_devfs()) {
-	sprintf(tem2,"/dev/ide/host%d/bus%d/target%d/lun0/", host, bus, target);
+	sprintf(tem2, DEV_DIR "/ide/host%d/bus%d/target%d/lun0/", host, bus, target);
 	strcat(strcpy(format,tem2), "part%d");
 	disc = strcat(tem2, "disc");
     }
@@ -313,7 +320,7 @@ void cache_ide (char *name, int major)
 
 void preload_dev_cache(void)
 {
-    char tmp[80];
+    char tmp[PATH_MAX];
     int i;
     int vsave;
 
@@ -324,54 +331,54 @@ void preload_dev_cache(void)
 #endif
 
     if (is_devfs()) {
-	cache_add("/dev/floppy/0", 0x0200);
-	cache_add("/dev/floppy/1", 0x0201);
+	cache_add(DEV_DIR "/floppy/0", 0x0200);
+	cache_add(DEV_DIR "/floppy/1", 0x0201);
     }
     else {
-	cache_add("/dev/fd0", 0x0200);
-	cache_add("/dev/fd1", 0x0201);
+	cache_add(DEV_DIR "/fd0", 0x0200);
+	cache_add(DEV_DIR "/fd1", 0x0201);
     }
     
 #if 1
-    cache_ide("/dev/hdt", MAJOR_IDE10);
-    cache_ide("/dev/hds", MAJOR_IDE10);
-    cache_ide("/dev/hdr", MAJOR_IDE9);
-    cache_ide("/dev/hdq", MAJOR_IDE9);
+    cache_ide(DEV_DISK_DIR "/hdt", MAJOR_IDE10);
+    cache_ide(DEV_DISK_DIR "/hds", MAJOR_IDE10);
+    cache_ide(DEV_DISK_DIR "/hdr", MAJOR_IDE9);
+    cache_ide(DEV_DISK_DIR "/hdq", MAJOR_IDE9);
 
-    cache_ide("/dev/hdp", MAJOR_IDE8);
-    cache_ide("/dev/hdo", MAJOR_IDE8);
-    cache_ide("/dev/hdn", MAJOR_IDE7);
-    cache_ide("/dev/hdm", MAJOR_IDE7);
+    cache_ide(DEV_DISK_DIR "/hdp", MAJOR_IDE8);
+    cache_ide(DEV_DISK_DIR "/hdo", MAJOR_IDE8);
+    cache_ide(DEV_DISK_DIR "/hdn", MAJOR_IDE7);
+    cache_ide(DEV_DISK_DIR "/hdm", MAJOR_IDE7);
     
-    cache_ide("/dev/hdl", MAJOR_IDE6);
-    cache_ide("/dev/hdk", MAJOR_IDE6);
+    cache_ide(DEV_DISK_DIR "/hdl", MAJOR_IDE6);
+    cache_ide(DEV_DISK_DIR "/hdk", MAJOR_IDE6);
 #ifdef MAJOR_IDE5
-    cache_ide("/dev/hdj", MAJOR_IDE5);
-    cache_ide("/dev/hdi", MAJOR_IDE5);
+    cache_ide(DEV_DISK_DIR "/hdj", MAJOR_IDE5);
+    cache_ide(DEV_DISK_DIR "/hdi", MAJOR_IDE5);
 #endif
 #endif
     
     for (i = MAX; i >= 0; i--) {
-	sprintf(tmp, is_devfs() ? "/dev/md/%d" : "/dev/md%d", i);
+	sprintf(tmp, is_devfs() ? DEV_DISK_DIR "/md/%d" : DEV_DISK_DIR "/md%d", i);
 	cache_add(tmp, MKDEV(MAJOR_MD,i));
     }
 
-    if (!is_devfs()) cache_scsi("/dev/sda", MAJOR_SD);
+    if (!is_devfs()) cache_scsi(DEV_DISK_DIR "/sda", MAJOR_SD);
     
-    cache_ide("/dev/hdh", MAJOR_IDE4);
-    cache_ide("/dev/hdg", MAJOR_IDE4);
-    cache_ide("/dev/hdf", MAJOR_IDE3);
-    cache_ide("/dev/hde", MAJOR_IDE3);
+    cache_ide(DEV_DISK_DIR "/hdh", MAJOR_IDE4);
+    cache_ide(DEV_DISK_DIR "/hdg", MAJOR_IDE4);
+    cache_ide(DEV_DISK_DIR "/hdf", MAJOR_IDE3);
+    cache_ide(DEV_DISK_DIR "/hde", MAJOR_IDE3);
 
     for (i = 0; i <= 7; i++) {
-	sprintf(tmp, is_devfs() ? "/dev/loop/%d" : "/dev/loop%d", i);
+	sprintf(tmp, is_devfs() ? DEV_DIR "/loop/%d" : DEV_DIR "/loop%d", i);
 	cache_add(tmp,0x700+i);
     }
 
-    cache_ide("/dev/hdd", MAJOR_IDE2);
-    cache_ide("/dev/hdc", MAJOR_IDE2);
-    cache_ide("/dev/hdb", MAJOR_HD);
-    cache_ide("/dev/hda", MAJOR_HD);
+    cache_ide(DEV_DISK_DIR "/hdd", MAJOR_IDE2);
+    cache_ide(DEV_DISK_DIR "/hdc", MAJOR_IDE2);
+    cache_ide(DEV_DISK_DIR "/hdb", MAJOR_HD);
+    cache_ide(DEV_DISK_DIR "/hda", MAJOR_HD);
     
     verbose = vsave;
 }
@@ -379,7 +386,7 @@ void preload_dev_cache(void)
 #undef MAX
 
 #define NDEVICE 256
-static unsigned short idevices[NDEVICE];
+static unsigned int idevices[NDEVICE];
 static int maxdev = 0;
 
 /* return 0 if device has not been backed up,		*/
@@ -404,10 +411,10 @@ static int dev_listed(unsigned short dev)
 /* 0 if no timestamp returned, and no backup file created */
 
 int make_backup(char *backup_file, int force_backup, BOOT_SECTOR *bsect,
-	int device, char *id)
+	unsigned int device, char *id)
 {
     struct stat st;
-    char temp_name[PATH_MAX+1];
+    char temp_name[PATH_MAX];
     int bck_file;
     int timestamp=0;
     char *filename = "boot";
@@ -542,7 +549,7 @@ static int inited = 0;
     if (!inited) {
 	struct stat st;
 	int fd, random;
-#define RANDOM "/dev/urandom"	
+#define RANDOM DEV_DIR "/urandom"	
 
 	inited = time(NULL);
 	if ( stat(RANDOM, &st)==0 && S_ISCHR(st.st_mode)
@@ -655,13 +662,7 @@ unsigned int register_bios(int bios, int device)
 	    
 	    dev_open(&dev, device_code[bios], O_BYPASS);
 	    dev_open(&deva, device, O_BYPASS);
-#if 0
-	    fprintf(errstd,
-	    	"\nLILO is unable to correctly determine the device codes assigned to two\n"
-	    	"of your disks by the System BIOS.  Consequently, it seems to have guessed\n"
-	    	"wrong.  The following conflict needs to be resolved with explicit\n"
-		"    \"disk=/dev/...  bios=0x\?\?\" lines in '%s'.\n\n", DFL_CONFIG);	    
-#endif
+
 	    die("Bios device code 0x%02X is being used by two disks\n\t%s (0x%04X)  and  %s (0x%04X)",
 	    	bios|0x80, dev.name, device_code[bios], deva.name, device);
 	}
@@ -753,7 +754,7 @@ enum {INVALID=1, DUPLICATE=2, REGENERATE=3, NTCAUTION=4};
 
 struct VolumeMgmt {
     char *name;		/* the name of the disk; e.g.  "/dev/hda" */
-    int device;		/* the device number (major, minor) */
+    unsigned int device;	/* the device number (major, minor) */
     int sort;		/* the device number used for sorting */
     int flag;		/* various flag bits */
     char nt[PART_MAX];	/* flag partitions which might be NT */
@@ -770,42 +771,46 @@ struct VolumeMgmt {
 };
 
 
-#ifdef LCF_ATARAID
-/* return boolean if VM is a component drive of ATA */
+#ifdef LCF_MDPRAID
+/* return boolean if VM is a component drive of MDP-RAID */
 
-static int is_ata(struct VolumeMgmt *vm, struct VolumeMgmt *ata)
+static int is_mdp(struct VolumeMgmt *vm, struct VolumeMgmt *mdp)
 {
-    int ata_fd;
+    int mdp_fd;
     DEVICE dev;
-    struct ata_array_info_s raid;
-    struct ata_version_s version;
+    struct md_version md_version_info;
+    md_array_info_t raid;
     int ret=0;
+
+    if (verbose>=2) printf("is_mdp:   %04X : %04X\n",
+          vm->device, mdp->device);
     
-    if ((ata_fd=dev_open(&dev, ata->device, O_NOACCESS) ) < 0)
-	die("Unable to open %s", ata->name);
+    if ((mdp_fd=dev_open(&dev, mdp->device, O_NOACCESS) ) < 0)
+	die("Unable to open %s", mdp->name);
 
-    if (ioctl(ata_fd, ATA_RAID_VERSION, &version) < 0) ret = -1;
-    else if (version.major != ATA_MAJOR_VERSION ||
-		version.minor < ATA_MINOR_VERSION ) {
+    if (ioctl(mdp_fd, RAID_VERSION, &md_version_info) < 0) ret = -1;
+    else if (md_version_info.major > 0  || 
+	                  md_version_info.minor < 90) {
+	
 	ret = -2;
-	fprintf(errstd, "Warning: "
-	    "ATAraid driver version (kernel %d.%d.%d) <> (lilo %d.%d.%d)",
-		version.major, version.minor, version.patchlevel,
-		ATA_MAJOR_VERSION, ATA_MINOR_VERSION, ATA_PATCHLEVEL );
-
+	warn("RAID versions other than 0.90 are not supported");
     }
-    else if (ioctl(ata_fd, ATA_GET_ARRAY_INFO, &raid) < 0) ret = -1;
-    else {
-	struct ata_disk_info_s disk;
+    else if (ioctl(mdp_fd, GET_ARRAY_INFO, &raid) < 0) ret = -1;
+    else if (raid.level == 1) {
+        md_disk_info_t disk;
 	int i;
+	
 	for (i=0; ret==0 && i<raid.active_disks; i++) {
 	    disk.number = i;
-	    if (ioctl(ata_fd, ATA_GET_DISK_INFO, &disk) < 0) ret = -1;
+	    if (ioctl(mdp_fd, GET_DISK_INFO, &disk) < 0) ret = -1;
 	    if (vm->device == MKDEV(disk.major, disk.minor)) ret = 1;
 	}
     }
+    /* else ret = 0;  already */
     dev_close(&dev);
 
+    if (verbose>=2) printf("is_mdp: returns %d\n", ret);
+    
     return ret;
 }
 #endif
@@ -844,9 +849,18 @@ static int winnt_check(struct VolumeMgmt *vm, int fatal)
 
 
 
+#ifdef LCF_MDPRAID
 #define SORT(d) (MAJOR(d)==MAJOR_SD?MKDEV(MAJOR_SD_SORT,MINOR(d)):\
+		 MAJOR(d)==MAJOR_MDP?MKDEV(MAJOR_MDP_SORT,MINOR(d)):\
+		 MAJOR(d)==MAJOR_EMD?MKDEV(MAJOR_EMD_SORT,MINOR(d)):\
 		 MAJOR(d)==MAJOR_HPT370?MKDEV(MAJOR_HPT370_SORT,MINOR(d)):\
 		 (d))
+#else
+#define SORT(d) (MAJOR(d)==MAJOR_SD?MKDEV(MAJOR_SD_SORT,MINOR(d)):\
+		 MAJOR(d)==MAJOR_EMD?MKDEV(MAJOR_EMD_SORT,MINOR(d)):\
+		 MAJOR(d)==MAJOR_HPT370?MKDEV(MAJOR_HPT370_SORT,MINOR(d)):\
+		 (d))
+#endif
 
 int pf_hard_disk_scan(void)
 {
@@ -854,12 +868,14 @@ int pf_hard_disk_scan(void)
 static int warned = 0, called = 0;
     char *line, *next, *name;
     int major, minor, i, ipart;
-    int dev, device, mask, bios;
+    int dev, mask, bios;
+    unsigned int device;
     size_t n;
     DEVICE Dev;
     DT_ENTRY *walk;
     struct stat st;
     int duplicate = 0, invalid = 0, ret = 0, ntcaution = 0;
+    int raidcaution = 0;
     long codes = 0L;
 
 /* called from  raid_setup  &  from  geo_open */
@@ -873,7 +889,7 @@ static int warned = 0, called = 0;
 #else
     if ((pp_fd = fopen(PARTITIONS, "r"))==NULL || fetch()) {
 #endif
-	fprintf(errstd, "Warning: '" PARTITIONS "' does not exist, disk scan bypassed\n");
+	warn("'" PARTITIONS "' does not exist, disk scan bypassed");
 	return 1;
     }
 
@@ -889,6 +905,25 @@ static int warned = 0, called = 0;
 
 	major = strtoul(line, &next, 10);
 	if (major==0 || line==next) continue;
+
+        if (is_dm_major(major)) {
+#ifndef LCF_DEVMAPPER
+            warn("device-mapper (%d) referenced in " PARTITIONS ",\n"
+                 "   but LILO is configured without DEVMAPPER option.  Skipping device.", major);
+            continue;
+#endif
+        }
+        else if ((major>=60 && major<=63) || (major>=120 && major<=127) ) {
+            warn(PARTITIONS " references Experimental major device %d.", major);
+        }
+        else if (major==255) {
+            warn(PARTITIONS " references Reserved device 255.");
+            continue;
+        }
+        else if (major>=240 && major<255) {
+            warn(PARTITIONS " references Experimental major device %d.", major);
+        }
+
 	minor = strtoul(next, &name, 10);
 	if (next==name) continue;
 	/* skip */ strtoull(name,&next,10);
@@ -897,10 +932,10 @@ static int warned = 0, called = 0;
 	name = next;
 	while (*name && !isspace(*name)) name++;
 	*name = 0;
-	if (strncmp("/dev/", next, 5) != 0) name = next-5;
+	if (strncmp(DEV_DISK_DIR "/", next, strlen(DEV_DISK_DIR)+1) != 0) name = next-(strlen(DEV_DISK_DIR)+1);
 	else name = next;
 	if (*name=='/') name++;
-	strncpy(name, "/dev/", 5);
+	strncpy(name, DEV_DISK_DIR "/", strlen(DEV_DISK_DIR)+1);
 	if (verbose>=5) {
 	    printf("pf_hard_disk_scan: (%d,%d) %s\n", major, minor, name);
 	}
@@ -910,29 +945,28 @@ static int warned = 0, called = 0;
 	Dev.delete = 0;
 	if (stat(name, &st) < 0) {
 	    dev_open(&Dev, device, O_BYPASS);
-	    if (!warned && !nowarn) {
-		fprintf(errstd,"Warning: '" PARTITIONS "' does not match '/dev' directory structure.\n"
-				"    Name change: '%s' -> '%s'\n%s"
+	    if (!warned) {
+		warn("'" PARTITIONS "' does not match '" DEV_DISK_DIR "' directory structure.\n"
+				"    Name change: '%s' -> '%s'%s"
 				,	name, Dev.name,
-				slashes(name) > 3 && slashes(Dev.name) < 3 ?
+				slashes(name) > 3 && slashes(Dev.name) < 3 ? "\n"
 				"    The kernel was compiled with DEVFS_FS, but 'devfs=mount' was omitted\n"
-				"        as a kernel command-line boot parameter; hence, the '/dev' directory\n"
-				"        structure does not reflect DEVFS_FS device names.\n"
-			:	slashes(name) < 3 && slashes(Dev.name) > 3 ?
-				"    The kernel was compiled without DEVFS, but the '/dev' directory structure\n"
-				"        implements the DEVFS filesystem.\n"
+				"        as a kernel command-line boot parameter; hence, the '" DEV_DISK_DIR "' directory\n"
+				"        structure does not reflect DEVFS_FS device names."
+			:	slashes(name) < 3 && slashes(Dev.name) > 3 ? "\n"
+				"    The kernel was compiled without DEVFS, but the '" DEV_DISK_DIR "' directory structure\n"
+				"        implements the DEVFS filesystem."
 			:
 				""
 				);
 		warned++;
 	    }
-	    else if (verbose >= 1 && !nowarn) {
-		fprintf(errstd,"    Name change: '%s' -> '%s'\n", name, Dev.name);
+	    else {
+		warn("Name change: '%s' -> '%s'", name, Dev.name);
 	    }
 	    name = Dev.name;
 	    if (Dev.delete) {
-	        if (!nowarn) fprintf(errstd, 
-		    "Warning: '/dev' directory structure is incomplete; device (%d, %d) is missing.\n",
+	        warn("'" DEV_DISK_DIR "' directory structure is incomplete; device (%d, %d) is missing.",
 	    			major, minor);
 	    	cache_add(name, device);
 	    }
@@ -947,8 +981,8 @@ static int warned = 0, called = 0;
 	for (walk=disktab; walk; walk=walk->next) {
 	    if (walk->device == dev) {
 		if (walk->heads == 0 /* inaccessible */) {
-		    if ((!nowarn || verbose>=2) && ipart==0 && !identify)  fprintf(errstd,
-			"Warning: bypassing VolumeID scan of drive flagged INACCESSIBLE:  %s\n", name);
+		    if (ipart==0 && !identify)
+		        warn("bypassing VolumeID scan of drive flagged INACCESSIBLE:  %s", name);
 		    ipart = -1;	/* causes skip below */
 		}
 		break;
@@ -978,8 +1012,8 @@ static int warned = 0, called = 0;
 		if (ndevs>=MAX_DEVICES) {
 		    die("More than %d hard disks are listed in '" PARTITIONS "'.\n"
 			"    Disks beyond the %dth must be marked:\n"
-			"        disk=/dev/XXXX  inaccessible\n"
-			"    in the configuration file (/etc/lilo.conf).\n"
+			"        disk=" DEV_DISK_DIR "/XXXX  inaccessible\n"
+			"    in the configuration file (" DFL_CONFIG ").\n"
 			, MAX_DEVICES, MAX_DEVICES);
 		}
 		else
@@ -1025,11 +1059,13 @@ static int warned = 0, called = 0;
 
 		    dev_close(&Dev2);
 
-#if 0
-		    geo_query_dev(&geo, dev, 1);
-#else
-		    geo_query_dev(&geo, dev, MAJOR(dev)!=MAJOR_LOOP);
+		    geo_query_dev(&geo, dev,
+		              MAJOR(dev)!=MAJOR_LOOP
+#ifdef LCF_ATARAID
+                               && MAJOR(dev)!=MAJOR_DM
 #endif
+		                                );
+
 		    vm[ndevs].bios.probe = bios_device(&geo, dev);
 
 		    if (serial_valid(serial, DEV_MASK)) {
@@ -1038,10 +1074,6 @@ static int warned = 0, called = 0;
 				duplicate++;
 				vm[i].flag |= DUPLICATE;
 				vm[ndevs].flag |= DUPLICATE;	/* mark both of them */
-#if 0
-				fprintf(errstd, "Error:  %s  and  %s  have duplicate volume IDs\n",
-					kname[i], name);
-#endif
 			    } /* if */
 			} /* for */
 		    }
@@ -1083,7 +1115,7 @@ static int warned = 0, called = 0;
     if (verbose>=2) printf("pf_hard_disk_scan: ndevs=%d\n", ndevs);
 
 /* now sort the volumes into canonical order */
-    {
+  {
 	int i,j,k;
 	
 	for (j=ndevs-1; j>0; j--)
@@ -1095,23 +1127,25 @@ static int warned = 0, called = 0;
 	    vm[i+1] = temp;
 	}
 
-/* now automatically treat ATARAID devices as inaccessible */
-      for (k=0; k<ndevs; k++)
-#ifdef LCF_ATARAID
-      if (MAJOR(vm[k].device) == MAJOR_HPT370) {
-#if ATAKEYWORD
-	if (!cfg_get_flag(cf_options, "ataraid") ) {
-	    if (!nowarn) fprintf(errstd,
-	    	"Warning: ATARAID controller present, without \"ataraid\" keyword used.\n"
-	    	"    Underlying drives individually must be marked INACCESSIBLE.\n" );
+/* now automatically treat MDP-RAID devices as inaccessible */
+    for (k=0; k<ndevs; k++) {
+#ifdef LCF_MDPRAID
+      if (MAJOR(vm[k].device) == MAJOR_MDP  ||
+          MAJOR(vm[k].device) == MAJOR_EMD   ) {
+
+        if (verbose>=2) printf("MDP-RAID detected,   k=%d\n", k);
+	if (cfg_get_flag(cf_options, "noraid") ) {
+	    raidcaution = 1;
+	    warn("RAID controller present, with \"noraid\" keyword used.\n"
+	    	"    Underlying drives individually must be marked INACCESSIBLE." );
 	} else {
-#endif
-	    for (j=k+1; j<ndevs; j++) {
-		if ((i=is_ata(&vm[j], &vm[k]))<0) {
-		    if (!nowarn && !identify) fprintf(errstd,
-			"Warning: (ATAraid driver) the kernel does not support underlying\n"
+	    for (j=0; j<ndevs; j++) {
+	        if (j==k) break;	/* skip ourselves */
+
+		if ((i=is_mdp(&vm[j], &vm[k]))<0) {
+		    if (!identify) warn("(MDP-RAID driver) the kernel does not support underlying\n"
 			"    device inquiries.  Each underlying drive of  %s  must\n"
-			"    individually be marked INACCESSIBLE.\n", vm[k].name
+			"    individually be marked INACCESSIBLE.", vm[k].name
 			);
 		    j=ndevs;	/* terminate loop */
 		}
@@ -1129,44 +1163,49 @@ static int warned = 0, called = 0;
 
 		    if (vm[j].dt->heads != 0) {
 			vm[j].dt->heads = 0;
-			if (!nowarn || verbose>=1) fprintf(errstd,
-				"Warning: (ataraid) underlying device flagged INACCESSIBLE: %s\n",
+			warn("(MDP-RAID) underlying device flagged INACCESSIBLE: %s",
 				vm[j].name);
 		    }
 		    --ndevs;
-		    if (!nowarn || verbose>=1)
-		    	fprintf(errstd,
-		    	"Warning: bypassing VolumeID check of underlying ATARAID drive:\n"
-		    	"\t%04X  %08X  %s\n",
+		    warn("bypassing VolumeID check of underlying MDP-RAID drive:\n"
+		    	"\t%04X  %08X  %s",
 		    		vm[j].device, vm[j].vol_id.kernel, vm[j].name);
 		    for (i=j; i<ndevs; i++) vm[i] = vm[i+1];
-		    j = k;	/* proper place to resume scan */
+		    if (j < k) k--;
+		    j--;
 		}
-	    }
-#if ATAKEYWORD
+	    } /* for j   ... */
 	}
-#endif
       }
 #else
-	if (MAJOR(vm[k].device) == MAJOR_HPT370) {
-	    if (!nowarn) fprintf(errstd,
-	    	"Warning: ATARAID controller present; underlying drives individually\n"
-	    	"    must be marked INACCESSIBLE.\n" );
+      if (MAJOR(vm[k].device) == MAJOR_EMD  ||
+          MAJOR(vm[k].device) == MAJOR_MDP   ) {
+          raidcaution = 1;
+          warn("MDP-RAID controller present; underlying drives individually\n"
+	    	"    must be marked INACCESSIBLE." );
 #if BETA_TEST
 {
-	int ata_fd;
+	int mdp_fd;
 	DEVICE dev;
 
-	if ((ata_fd=dev_open(&dev, vm[k].device, O_NOACCESS) ) < 0)
+	if ((mdp_fd=dev_open(&dev, vm[k].device, O_NOACCESS) ) < 0)
 	    die("Unable to open %s",vm[k].name);
-	if (ioctl(ata_fd, ATA_PRINT_RAID_DEBUG, NULL) < 0)
-	    die("ATA_PRINT_RAID_DEBUG not implemented");
+
 	dev_close(&dev);
 }
 #endif
-	}
+      }
 #endif
-    } /* end sort */
+#ifdef LCF_ATARAID
+      if ( MAJOR(vm[k].device) == MAJOR_DM ) {
+          if (verbose>=2) printf("ATA-RAID detected,   k=%d\n", k);
+          raidcaution = 1;
+          warn("ATA-RAID controller present;\n"
+              "    Underlying drives individually must be marked INACCESSIBLE." );
+      }
+#endif
+    } /* for (k ... */
+  } /* end sort */
 
 	if (verbose>=3) {
 	    for (i=0; i<ndevs; i++)
@@ -1194,6 +1233,15 @@ static int warned = 0, called = 0;
 
     while (duplicate) {		/* loop until there are none */
 	int j, k;
+	
+	if (raidcaution) {
+	    raidcaution = 0;	/* print comment only once */
+	    warn("Duplicated VolumeID's will be overwritten;\n"
+	        "   With RAID present, this may defeat all boot redundancy.\n"
+	        "   Underlying RAID-1 drives should be marked INACCESSIBLE.\n"
+	        "   Check 'man lilo.conf' under 'disk=', 'inaccessible' option."
+	        );
+	}
 	duplicate = 0;
 	for (j=ndevs-1; j>0; j--) {
 	    for (i=0; i<j; i++) {
@@ -1305,14 +1353,12 @@ static int warned = 0, called = 0;
 	}
 	else {
 	    vm[i].dt->heads = 0;	/* mark inaccessible */
-	    if (!nowarn) {
+	    {
 		static int i=0;
-		if (!(i++)) fprintf(errstd,
-		    "Warning: Internal implementation restriction. Boot may occur from the first\n"
-		    "    %d disks only. Disks beyond the %dth will be flagged INACCESSIBLE.\n"
+		if (!(i++)) warn("Internal implementation restriction. Boot may occur from the first\n"
+		    "    %d disks only. Disks beyond the %dth will be flagged INACCESSIBLE."
 		    , MAX_BIOS_DEVICES, MAX_BIOS_DEVICES);
-		fprintf(errstd,
-		"Warning: 'disk=%s  inaccessible' is being assumed.  (%04X)\n",
+		warn("'disk=%s  inaccessible' is being assumed.  (%04X)",
 		vm[i].name, vm[i].device);
 	    }
 	}
@@ -1324,10 +1370,10 @@ static int warned = 0, called = 0;
 
     for (bios=8*sizeof(codes)-1; !(codes&(1L<<bios)) && bios>=0; ) bios--;
 
-    if (bios > ndevs && !nowarn)
-	fprintf(errstd,"Warning: BIOS device code 0x%02X is used (>0x%02X).  It indicates more disks\n"
+    if (bios > ndevs)
+	warn("BIOS device code 0x%02X is used (>0x%02X).  It indicates more disks\n"
 			"  than those represented in '/proc/partitions' having actual partitions.\n"
-			"  Booting results may be unpredictable.\n", bios+0x80, ndevs+0x80-1);
+			"  Booting results may be unpredictable.", bios+0x80, ndevs+0x80-1);
 
 
 
