@@ -1,6 +1,6 @@
 /* test4.c */
 /*
- Copyright 2001-2004 John Coffman.
+ Copyright 2001-2005 John Coffman.
  All rights reserved.
 
  Licensed under the terms contained in the file 'COPYING' in the LILO
@@ -10,7 +10,7 @@
 #include <bios.h>
 #include "../bdata.h"
 
-#define DISK_VERSION "2.3"
+#define DISK_VERSION "2.4"
 
 #if __MSDOS__==0
 #define putch bios_putc
@@ -28,6 +28,8 @@
 #define SPACER "\f\n"
 #define SEQ 0x3C4
 
+#define FS_REG 0x80
+
 #ifndef EDD_SUBSET
 #define EDD_SUBSET 4
 #define EDD_LOCK 2
@@ -49,6 +51,7 @@ union REGS reg, oreg;
 struct SREGS sreg;
 int last_good_disk;
 int video_1 = 0xF00;
+int fs_mod = 0;
 int num_hd = BD_MAX_HARD;
 int errno;
 dword hma;	/* highest memory address */
@@ -70,6 +73,41 @@ struct gdt_entry {
 
 struct gdt_entry gdt[6];
 
+static
+int get_fs(void)
+{
+#asm
+ mov ax,fs
+#endasm
+}
+
+static
+int set_fs(int val)
+{
+    int i = val;
+#asm
+ mov ax,4[bp]
+ mov fs,ax
+#endasm
+    return i;
+}
+
+static
+int check_fs(void)
+{
+    int ret = 0;
+
+#if DEBUG>=2    
+    printf("#");
+#endif    
+    if (get_fs() != FS_REG) {
+        fs_mod = ret = 1;
+        printf("\nThe FS register has been modified.\n");
+        set_fs(FS_REG);
+    }
+    return ret;
+}
+
 static
 int hicopy (unsigned long to, unsigned long from, int wcount)
 {
@@ -249,7 +287,7 @@ void banner(char *version)
 {
 	printf(	"\n\n\n"
 ">>>> Disk Detection and Parameter Display <<<<\n\n\n"
-"Version %s, Copyright (C) 1999-2004  John Coffman <johninsd@san.rr.com>\n"
+"Version %s, Copyright (C) 1999-2005  John Coffman <johninsd@san.rr.com>\n"
 "Portions Copyright (C) 1996-2001 Robert de Bath, used with permission\n"
 "Re-use and redistribution rights set forth in the file \"COPYING\".\n\n",
 	 version);
@@ -400,7 +438,8 @@ static
 void pause(void)
 {
 	char ch;
-/* Must be standalone */    
+/* Must be standalone */
+        check_fs();
 	printf("\n\n\nHit <Enter> to continue, <^C> to quit ...");
 	do {
 	    ch = getch();
@@ -999,6 +1038,7 @@ static char *dt[] = {	"5.25\", 40 track, 360K",
 			printf("Error:  Hard disk BIOS should not touch ES:DI\n");
 		}
 		s = (oreg.h.cl & 0x3F);		/* sectors 1..63 */
+		if (s == 0) s = 64;
 		h = (int)oreg.h.dh + 1;		/* heads 0..254 */
 		c = (((oreg.h.cl & 0xC0)<<2) | oreg.h.ch) + 1;
 		printf("Disk geometry (C:H:S) = %d:%d:%d (", c, h, s);
@@ -1007,6 +1047,8 @@ static char *dt[] = {	"5.25\", 40 track, 360K",
 		sizeit(sect);
 		printf("\n%s disks on system = %d\n",
 			dev&0x80 ? "Fixed" : "Floppy", oreg.h.dl);
+                if (s > 63)  printf("BIOS BUG!!!  sectors returned as zero; 64 assumed\n");
+                if (h > 255) printf("BIOS BUG!!!  heads > 255; BIOS is not IBM compatible\n");
 		if (dev & 0x80) {
 		    if (dev == 0x80) num_hd = oreg.h.dl;
 		    do_edd(dev);
@@ -1050,6 +1092,7 @@ void do_get_pt(int dev)
     if (m) print_io_status(m);
     else printf("  okay");
     printf("\n");
+    check_fs();
 }
 
 static
@@ -1141,6 +1184,12 @@ void main(void)
 {
     int m, i, dev;
     
+    set_fs(FS_REG);
+#if DEBUG>=1
+    printf("FS=%04x\n", get_fs());
+    pause();
+#endif
+    
     if (!is_msdos()) {
 /**	atexit(pause);	**/
 	video_fix();	/* for Dumb DELL computers */
@@ -1154,9 +1203,12 @@ void main(void)
         
     segread(&sreg);
     printf("\nBeginning of '_main'\n");
+    check_fs();
     print_sregs(&sreg);
 #endif
+    check_fs();
     banner(DISK_VERSION);
+    check_fs();
     v86test();
     testDX();
 #if DEBUG>=2
@@ -1196,6 +1248,8 @@ void main(void)
     
     if (!is_msdos()) {
 	printf("\n\nInitial SEQ reg 1:  0x%02x\n", video_1);
-	pause();
     }
+    printf("The FS register was %smodified during the tests.\n",
+	    fs_mod ? "" : "NOT ");
+    pause();
 }
