@@ -15,6 +15,9 @@ source directory.
 #include <stdio.h>
 #include <fcntl.h>
 #include <errno.h>
+#include <stdlib.h>
+#include <limits.h>
+#include <ctype.h>
 #include <sys/types.h>
 
 #ifdef LCF_REISERFS
@@ -27,7 +30,11 @@ source directory.
 
 #include <string.h>
 
+
 #include "config.h"
+#ifdef LCF_DEVMAPPER
+# include <libdevmapper.h>
+#endif
 #include "lilo.h"
 #include "common.h"
 #include "device.h"
@@ -61,17 +68,43 @@ source directory.
 #endif
 #endif
 
+#ifdef LCF_DEVMAPPER
+typedef struct _dm_target {
+    struct _dm_target *next;
+    uint64_t start,length,offset;
+    int device;
+} DM_TARGET;
+
+typedef struct _dm_table {
+    struct _dm_table *next;
+    int device;
+    struct _dm_target *target;
+} DM_TABLE;
+
+DM_TABLE *dmtab = NULL;
+int dm_version_nr = 0;
+#endif
+
+int dm_major_list[16];
+int dm_major_nr;
+
 #ifdef LCF_LVM
 struct lv_bmap {
     __u32 lv_block;
-    __u16 lv_dev;
+    dev_t lv_dev;			/* was __u16, which is wrong */
 };
 
 #ifndef LV_BMAP
-#define LV_BMAP				_IOWR(0xfe, 0x30, char)
+/*	Should the definition be:					*/
+#define LV_BMAP				_IOWR(0xfe, 0x30, int)
+/*	As defined in the 2.4 kernels:					*/
+/*#define LV_BMAP				_IOWR(0xfe, 0x30, 1)  */
 #endif
 #ifndef LVM_GET_IOP_VERSION
-#define LVM_GET_IOP_VERSION		_IOR(0xfe, 0x98, char)
+/*	Should the definition be:					*/
+#define LVM_GET_IOP_VERSION		_IOR(0xfe, 0x98, unsigned short)
+/*	As defined in the 2.4 kernels:					*/
+/*#define LVM_GET_IOP_VERSION		_IOR(0xfe, 0x98, 1) */
 #endif
 #endif
 
@@ -113,8 +146,14 @@ int old_disktab = 0;
 
 void geo_init(char *name)
 {
-    FILE *file;
+    FILE *file = NULL;
     char line[MAX_LINE+1];
+#ifdef LCF_DEVMAPPER
+    struct dm_task *dmt;
+    char dm_version[32];
+#endif
+    char major_name[32];
+    int major;
     char *here;
     DT_ENTRY *entry;
     int disk_section,items;
@@ -123,32 +162,88 @@ void geo_init(char *name)
 	if ((file = fopen(name,"r")) == NULL)
 	    die("open %s: %s",name,strerror(errno));
     }
-    else if ((file = fopen(DFL_DISKTAB,"r")) == NULL) return;
-    disk_section = !!disktab;
-    while (fgets(line,MAX_LINE,file)) {
-	here = strchr(line,'\n');
-	if (here) *here = 0;
-	here = strchr(line,'#');
-	if (here) *here = 0;
-	if (strspn(line," \t") != strlen(line)) {
-	    entry = alloc_t(DT_ENTRY);
-	    items = sscanf(line,"0x%x 0x%x %d %d %d %d",&entry->device,
-	      &entry->bios,&entry->sectors,&entry->heads,&entry->cylinders,
-	      &entry->start);
-	    if (items == 5) entry->start = -1;
-	    if (items < 5)
-		die("Invalid line in %s:\n\"%s\"",name ? name : DFL_DISKTAB,
-		  line);
-	    entry->next = disktab;
-	    disktab = entry;
-	    if (disk_section) die("DISKTAB and DISK are mutually exclusive");
-	    old_disktab = 1;
+    if (name || (file = fopen(DFL_DISKTAB,"r")) != NULL) {
+	disk_section = !!disktab;
+	while (fgets(line,MAX_LINE,file)) {
+	    here = strchr(line,'\n');
+	    if (here) *here = 0;
+	    here = strchr(line,'#');
+	    if (here) *here = 0;
+	    if (strspn(line," \t") != strlen(line)) {
+		entry = alloc_t(DT_ENTRY);
+		items = sscanf(line,"0x%x 0x%x %d %d %d %d",&entry->device,
+		  &entry->bios,&entry->sectors,&entry->heads,&entry->cylinders,
+		  &entry->start);
+		if (items == 5) entry->start = -1;
+		if (items < 5)
+		    die("Invalid line in %s:\n\"%s\"",name ? name : DFL_DISKTAB,
+		      line);
+		entry->next = disktab;
+		disktab = entry;
+		if (disk_section) die("DISKTAB and DISK are mutually exclusive");
+		old_disktab = 1;
+	    }
 	}
+	(void) fclose(file);
     }
+
+    dm_major_nr = 0;
+    file = fopen("/proc/devices", "r");
+    if (!file) return;
+
+    do {
+	if (!fgets(line, (sizeof line)-1, file)) {
+	    (void) fclose(file);
+	    return;
+	}
+	line[(sizeof line)-1] = 0;
+    } while(strncmp(line, "Block", 5) != 0);
+
+    while(fgets(line, (sizeof line)-1, file)) {
+	if (sscanf(line, "%d %31s\n", &major, major_name) != 2) continue;
+	if (strcmp(major_name, "device-mapper") != 0) continue;
+	dm_major_list[dm_major_nr] = major;
+	if (verbose >= 3) {
+	    printf("device-mapper major = %d\n", major);
+	}
+	if (++dm_major_nr > nelem(dm_major_list) ) break;
+    }
+
     (void) fclose(file);
+
+
+#ifdef LCF_DEVMAPPER
+    if (!(dmt = dm_task_create(DM_DEVICE_VERSION)))
+	return;
+    if (!dm_task_run(dmt))
+	return;
+    if (!dm_task_get_driver_version(dmt, dm_version, sizeof dm_version))
+	return;
+
+    /*
+     * to not confuse returned device number formats %02x:%02x and %d:%d
+     * we assume here that the %02x:%02x format is only found in the ioctl
+     * interface version < 4 (this is really getting ugly...)
+     */
+    dm_version_nr = atoi(dm_version);
+
+    dm_task_destroy(dmt);
+#endif
 }
 
 
+int is_dm_major(int major)
+{
+    int i;
+    
+    for(i=0; i<dm_major_nr; i++) {
+        if (dm_major_list[i] == major) return 1;
+    }
+    return 0;
+}
+
+
+
 void do_partition(void)
 {
     DT_ENTRY *entry,*walk;
@@ -183,10 +278,8 @@ int has_partitions_beta(dev_t dev)
     if (
       major == MAJOR_HD || major == MAJOR_IDE2 ||
       major == MAJOR_IDE3 || major == MAJOR_IDE4 ||
-#ifdef MAJOR_IDE5
-      major == MAJOR_IDE5 || 
-#endif
-      major == MAJOR_IDE6 ||
+      major == MAJOR_IDE5 || major == MAJOR_IDE6 ||
+      major == MAJOR_EMD  ||
       (major >= MAJOR_IDE7 && major <= MAJOR_IDE10) ||
       major == MAJOR_XT || major == MAJOR_ESDI || major == MAJOR_ACORN
       	) return 0xFFFFFFC0;	/* 6 bit partition mask */
@@ -199,8 +292,7 @@ int has_partitions_beta(dev_t dev)
       (major >= MAJOR_SMART2 && major <= MAJOR_SMART2+7) ||
       (major >= MAJOR_CISS && major <= MAJOR_CISS+7) ||
       major == MAJOR_FTL || major == MAJOR_NFTL || major == MAJOR_DOC ||
-      (major >= MAJOR_SD9 && major <= MAJOR_SD16) ||
-      major == MAJOR_EMD
+      (major >= MAJOR_SD9 && major <= MAJOR_SD16)
         ) return 0xFFFFFFF0;	/* 4 bit partition mask */
 
     if ( major == MAJOR_SATA || major == MAJOR_SATA2
@@ -237,7 +329,7 @@ unsigned char max_partno[512] = {
   0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
 */
  15, 15, 15, 15, 15, 15, 15, 15,  7,  7,  7,  7,  7,  7,  7,  7,   /*  8x   */
-  0,  0,  0,  0,  0,  0,  0,  0,  0, 15,  0,  0,  0,  0,  0,  0,
+  0,  0,  0,  0,  0,  0,  0,  0,  0, 63,  0,  0,  0,  0,  0,  0,
  31, 31,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
 
@@ -245,7 +337,7 @@ unsigned char max_partno[512] = {
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
-/*
+/*								   
   0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
 */
 };
@@ -256,7 +348,7 @@ int has_partitions(dev_t dev)
    int ret=0;
    
    if (major >= nelem(max_partno)) {
-	fprintf(errstd,"Warning: Major Device (%d) > %d\n",
+	warn("Major Device (%d) > %d",
 		major, nelem(max_partno)-1);
    }
    else if ( (ret=max_partno[major]) )  ret ^= 0xFFFFFFFF;
@@ -315,12 +407,8 @@ void do_disk(void)
 	}
     if (cfg_get_flag(cf_disk,"inaccessible")) {
 	entry->heads = 0;
-	if (
-#if 0
-		cfg_get_strg(cf_disk,"bios") || 
-#endif
-		cfg_get_strg(cf_disk,"sectors") ||
-	  cfg_get_strg(cf_disk,"heads") || cfg_get_strg(cf_disk,"cylinders"))
+	if (bios) die("INACCESSIBLE and BIOS are mutually exclusive");
+	if ( sectors || heads || cylinders )
 	    die("No geometry variables allowed if INACCESSIBLE");
     }
     entry->cylinders = cylinders ? to_number(cylinders) : -1;
@@ -428,7 +516,7 @@ void lvm_bmap(struct lv_bmap *lbm)
     static dev_t last_dev = 0;
 
     if (lbm->lv_dev != last_dev) {
-	char lvm_char[] = "/dev/lvm";
+	char lvm_char[] = DEV_DIR "/lvm";
 	unsigned short iop;
 
 	if (lvmfd != -1)
@@ -465,10 +553,10 @@ void evms_bmap(struct evms_get_bmap_t *ebm)
     static dev_t evms_last_dev = 0;
 
     if (ebm->dev != evms_last_dev) {
-        char evms_blk[] = "/dev/evms/block_device";
+        char evms_blk[] = DEV_DIR "/evms/block_device";
         struct evms_version_t evms_ver;
 
-        // Open the EVMS device
+        /* Open the EVMS device */
         if (evms_fd != -1)
             close(evms_fd);
 
@@ -476,11 +564,11 @@ void evms_bmap(struct evms_get_bmap_t *ebm)
         if (evms_fd < 0)
             die("Can't open EVMS block device %s.\n", evms_blk);
 
-        // Get EVMS ioctl version number.
+        /* Get EVMS ioctl version number. */
         if (ioctl(evms_fd, EVMS_GET_IOCTL_VERSION, &evms_ver) < 0)
             die("EVMS_GET_IOCTL_VERSION failed on %s.\n", evms_blk);
 
-        // Check that the ioctl version is >= 7.1.0
+        /* Check that the ioctl version is >= 7.1.0 */
         if (evms_ver.major < 7 ||
             (evms_ver.major == 7 && evms_ver.minor < 1))
             die("EVMS ioctl version %d.%d.%d does not support booting.\n",
@@ -591,6 +679,20 @@ void geo_query_dev(GEOMETRY *geo,int device,int all)
 	    geo->sectors = hdprm.sectors;
 	    geo->start = hdprm.start;
 	    break;
+	MASK31:
+	    geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 5);
+	    if (!get_all) break;
+	    if (ioctl(fd,HDIO_GETGEO,&hdprm) < 0)
+		die("geo_query_dev HDIO_GETGEO (dev 0x%04x): %s",device,
+		  strerror(errno));
+	    if (all && !hdprm.sectors)
+		die("HDIO_REQ not supported for your Disk controller. Please "
+		  "use a DISK section");
+	    geo->heads = hdprm.heads;
+	    geo->cylinders = hdprm.cylinders;
+	    geo->sectors = hdprm.sectors;
+	    geo->start = hdprm.start;
+	    break;
 	case MAJOR_DAC960:
 	case MAJOR_DAC960+1:
 	case MAJOR_DAC960+2:
@@ -664,10 +766,12 @@ void geo_query_dev(GEOMETRY *geo,int device,int all)
 	default:
 	    if (max_partno[major] && major==MAJOR_LOOP) break;
 	    if (max_partno[major] == 63)  goto MASK63;
+	    if (max_partno[major] == 31)  goto MASK31;
 	    if (max_partno[major] == 15)  goto MASK15;
 	    if (max_partno[major] == 7)   goto MASK7;
 	    
-	    if (MAJOR(device)>=120 && MAJOR(device)<=127)
+	    if ((MAJOR(device)>=120 && MAJOR(device)<=127)  ||
+	        (MAJOR(device)>=240 && MAJOR(device)<=254) )
 		die("Linux experimental device 0x04x needs to be defined.\n"
 		    "Check 'man lilo.conf' under 'disk=' and 'max-partitions='", device);
 	    else die("Sorry, don't know how to handle device 0x%04x",device);
@@ -774,6 +878,161 @@ void geo_get(GEOMETRY *geo,int device,int user_device,int all)
 {
     DT_ENTRY *walk;
     int inherited,keep_cyls,is_raid=0;
+#ifdef LCF_DEVMAPPER
+    int i;
+
+    for(i = 0; i < dm_major_nr; i++)
+	if (MAJOR(device) == dm_major_list[i])
+	    break;
+    while (i < dm_major_nr) {
+	DM_TABLE *dm_table;
+
+	for(dm_table = dmtab; dm_table; dm_table = dm_table->next)
+	    if (dm_table->device == device)
+		break;
+
+	if (dm_table) {
+	    DM_TARGET *target;
+
+	     device = 0;
+	    for(target = dm_table->target; target; target = target->next)
+		device = target->device;
+	} else {
+	    DEVICE dev;
+	    struct dm_task *dmt;
+	    void *next = NULL;
+	    char dmdev[PATH_MAX+1];
+	    char buf[PATH_MAX+1];
+	    char *slash;
+	    int result;
+
+	    dev_open(&dev, device, -1);
+	    strncpy(dmdev, dev.name, PATH_MAX);
+	    dmdev[PATH_MAX] = 0;
+	    do {
+		memset(buf, 0, PATH_MAX + 1);
+		if ((result = readlink(dmdev, buf, PATH_MAX)) < 0 && errno != EINVAL)
+		    die("device-mapper: readlink(\"%s\") failed with: %s",buf,
+			strerror(errno));
+		if (result >= 0) {
+		    if (buf[0] != '/' && (slash = strrchr(dmdev, '/')) != NULL)
+			slash++;
+		    else
+			slash = dmdev;
+		    strncpy(slash, buf, PATH_MAX - (slash-dmdev));
+		}
+		if (realpath(dmdev, buf) == NULL)
+		    die("device-mapper: realpath(\"%s\") failed with: %s",dmdev,
+			strerror(errno));
+		strncpy(dmdev, buf, PATH_MAX);
+	    } while (result >= 0);
+	    dmdev[PATH_MAX] = 0;
+
+	    if (!(dmt = dm_task_create(DM_DEVICE_TABLE)))
+		die("device-mapper: dm_task_create(DM_DEVICE_TABLE) failed");
+	    slash = strrchr(dmdev, '/');
+		if (slash)
+	    slash++;
+		else
+	    slash = dmdev;
+	    if (!dm_task_set_name(dmt, slash))
+		die("device-mapper: dm_task_set_name(\"%s\") failed",dmdev);
+	    if (!dm_task_run(dmt))
+		die("device-mapper: dm_task_run(DM_DEVICE_TABLE) failed");
+
+	    dm_table = alloc_t(DM_TABLE);
+	    dm_table->device = device;
+	    dm_table->target = NULL;
+	    dm_table->next = dmtab;
+	    dmtab = dm_table;
+
+	    device = 0;
+
+	    do {
+		DM_TARGET *target;
+		uint64_t start,length;
+		int major,minor;
+		char *target_type,*params;
+		char *p;
+
+		next = dm_get_next_target(dmt, next, &start, &length,
+		  &target_type, &params);
+
+		if (!target_type) continue;
+
+		if (strcmp(target_type, "linear") != 0)
+		    die("device-mapper: only linear boot device supported");
+
+		target = alloc_t(DM_TARGET);
+		target->start = start;
+		target->length = length;
+		if (dm_version_nr < 4 &&
+		    isxdigit(params[0]) &&
+		    isxdigit(params[1]) &&
+		    params[2] == ':' &&
+		    isxdigit(params[3]) &&
+		    isxdigit(params[4])) { /* old 2.4 format */
+		    if (sscanf(params, "%02x:%02x %"PRIu64, &major, &minor, &target->offset) != 3)
+			die("device-mapper: parse error in linear params (\"%s\")", params);
+		} else if (isdigit(params[0]) &&
+			   strchr(params, ':')) { /* dm_bdevname/format_dev_t (>= 2.6.0-test4?) format */
+		    if (sscanf(params, "%u:%u %"PRIu64, &major, &minor, &target->offset) != 3)
+			die("device-mapper: parse error in linear params (\"%s\")", params);
+		} else { /* >= 2.5.69 format, this should go away soon */
+		    struct stat st;
+		    FILE *file;
+
+		    p = strrchr(params, ' ');
+		    if (p == NULL)
+			die("device-mapper: parse error in linear params (\"%s\")", params);
+		    *p = 0;
+		    sprintf(buf, DEV_DIR "/%s", params);	/* let's hope it's there */
+		    if (stat(buf, &st) == 0) {
+			if (!S_ISBLK(st.st_mode))
+			    die("device-mapper: %s is not a valid block device", buf);
+			major = MAJOR(st.st_rdev);
+			minor = MINOR(st.st_rdev);
+		    } else {				/* let's try sysfs */
+			int dev;
+			sprintf(buf, "/sys/block/%s/dev", params);
+			file = fopen(buf, "r");
+ 			if (!file)
+			    die("device-mapper: \"%s\" could not be opened. /sys mounted?", buf);
+			if (!fgets(buf, PATH_MAX, file))
+			    die("device-mapper: read error from \"/sys/block/%s/dev\"", params);
+			if (sscanf(buf, "%u:%u", &major, &minor) != 2) {
+			    if (sscanf(buf, "%x", &dev) != 1)
+				die("device-mapper: error getting device from \"%s\"", buf);
+			    major = MAJOR(dev);
+			    minor = MINOR(dev);
+			}
+			(void) fclose(file);
+		    }
+		    *p = ' ';
+		    if (sscanf(p+1, "%"PRIu64, &target->offset) != 1)
+			die("device-mapper: parse error in linear params (\"%s\")", params);
+		}
+		target->device = (major << 8) | minor;
+		if (!device)
+		    device = target->device;
+		target->next = dm_table->target;
+		dm_table->target = target;
+	    } while(next);
+
+	    dm_task_destroy(dmt);
+
+	    dev_close(&dev);
+	}
+
+	if (!device)
+	    die("device-mapper: Error finding real device");
+	geo->base_dev = device;
+
+	for(i = 0; i < dm_major_nr; i++)
+	    if (MAJOR(device) == dm_major_list[i])
+		break;
+    }
+#endif
 
     if (verbose>=5) printf("geo_get: device %04X, all=%d\n", device, all);
 #ifdef LCF_LVM
@@ -830,10 +1089,10 @@ void geo_get(GEOMETRY *geo,int device,int user_device,int all)
 	md_disk_info_t md_disk_info;
 	int raid_limit;
 
-	sprintf(mdxxx, "/dev/md%d", MINOR(device));
+	sprintf(mdxxx, DEV_DISK_DIR "/md%d", MINOR(device));
 	if ((md_fd=open(mdxxx,O_NOACCESS)) < 0)
 	{
-	    sprintf(mdxxx, "/dev/md/%d", MINOR(device));
+	    sprintf(mdxxx, DEV_DIR "/md/%d", MINOR(device));
 	    if ((md_fd=open(mdxxx,O_NOACCESS)) < 0)
 		die("Unable to open %s", mdxxx);
 	}
@@ -940,9 +1199,9 @@ int pass;
 	
 	if (all) bios_device(geo, device);
 	
-	if (!nowarn && (geo->device & 0x7f) >= bios_max_devs() &&
+	if ((geo->device & 0x7f) >= bios_max_devs() &&
 	  user_device == -1 && (!walk || walk->bios == -1))
-	    fprintf(errstd,"Warning: BIOS drive 0x%02x may not be accessible\n",
+	    warn("BIOS drive 0x%02x may not be accessible",
 	      geo->device);
     }
     if (walk) {
@@ -966,18 +1225,21 @@ int pass;
     if (geo->heads > BIOS_MAX_HEADS)
 	die("Device 0x%04X: Maximum number of heads is %d, not %d\n",device,
 	  BIOS_MAX_HEADS,geo->heads);
-    if (geo->heads == BIOS_MAX_HEADS && !nowarn)
-	fprintf(errstd, "Warning: Maximum number of heads = %d (as specified)\n"
-			"   exceeds standard BIOS maximum of 255.\n", geo->heads);
+    if (geo->heads == BIOS_MAX_HEADS)
+	warn("Maximum number of heads = %d (as specified)\n"
+			"   exceeds standard BIOS maximum of 255.", geo->heads);
     if (geo->sectors > BIOS_MAX_SECS)
 	die("Device 0x%04X: Maximum number of sectors is %d, not %d\n",
 	  device,BIOS_MAX_SECS,geo->sectors);
+    if (geo->sectors == BIOS_MAX_SECS)
+        warn("Maximum number of heads = %d (as specified)\n"
+                        "   exceeds standard BIOS maximum of 63.", geo->sectors);
     if (!lba32 &&
-      (geo->start+geo->sectors-1)/geo->heads/geo->sectors >= BIOS_MAX_CYLS &&
-      !nowarn) {
-	fprintf(errstd,"Warning: device 0x%04x exceeds %d cylinder limit.\n"
-        "   Use of the 'lba32' option may help on newer (EDD BIOS) systems.\n",
-	  device,BIOS_MAX_CYLS);
+      (geo->start+geo->sectors-1)/geo->heads/geo->sectors >= BIOS_MAX_CYLS
+    ) {
+	warn("device 0x%04x exceeds %d cylinder limit.\n"
+        "   Use of the 'lba32' option may help on newer (EDD-BIOS) systems.",
+	  device, BIOS_MAX_CYLS);
     }
     if (verbose >= 3) {
 	printf("Device 0x%04x: BIOS drive 0x%02x, %d heads, %d cylinders,\n",
@@ -1004,9 +1266,9 @@ int geo_open(GEOMETRY *geo,char *name,int flags)
 
     if ((here = strrchr(name,':')) == NULL) user_dev = -1;
     else {
-	fprintf(errstd,":BIOS syntax is no longer supported. Please use a "
-	  "DISK section\n");
 	*here++ = 0;
+        warn("%s:BIOS syntax is no longer supported.\n    Please use a "
+	  "DISK section.", name);
 	user_dev = to_number(here);
     }
     if ((geo->fd = open(name,flags)) < 0)
@@ -1033,7 +1295,7 @@ int geo_open(GEOMETRY *geo,char *name,int flags)
     if (!geo->file) geo->spb = 2;
     else {
 	if (ioctl(geo->fd,FIGETBSZ,&block_size) < 0) {
-	    fprintf(errstd,"FIGETBSZ %s: %s\n",name,strerror(errno));
+	    warn("FIGETBSZ %s: %s",name,strerror(errno));
 	    geo->spb = 2;
 	}
 	else {
@@ -1091,7 +1353,7 @@ void geo_close(GEOMETRY *geo)
 int geo_comp_addr(GEOMETRY *geo,int offset,SECTOR_ADDR *addr)
 {
     int block,sector;
-    static int warnings = 0;
+    static int linear_warnings = 0;
 
 #if BETA_TEST
     if (verbose>=6)
@@ -1167,14 +1429,50 @@ int geo_comp_addr(GEOMETRY *geo,int offset,SECTOR_ADDR *addr)
             die("EVMS boot volume cannot be on multiple disks.\n");
         sector = ebm.rsector + ((offset/SECTOR_SIZE) % geo->spb) + geo->start;
     }
-    else {
-        sector = block*geo->spb+((offset/SECTOR_SIZE) % geo->spb);
-        sector += geo->start;
-    }
-#else
-    sector = block*geo->spb+((offset/SECTOR_SIZE) % geo->spb);
-    sector += geo->start;
+    else
 #endif
+    {
+#ifdef LCF_DEVMAPPER 
+	int dev = geo->dev;
+	int i;
+#endif
+	sector = block*geo->spb+((offset/SECTOR_SIZE) % geo->spb);
+#ifdef LCF_DEVMAPPER 
+	for(i = 0; i < dm_major_nr; i++)
+	    if (MAJOR(dev) == dm_major_list[i])
+		break;
+	while (i < dm_major_nr) {
+	    DM_TABLE *dm_table;
+	    DM_TARGET *dm_target;
+
+	    for(dm_table = dmtab; dm_table; dm_table = dm_table->next)
+		if (dm_table->device == dev)
+		    break;
+	    if (!dm_table)
+		die("device-mapper: Mapped device suddenly lost? (%d)", dev);
+
+	    for(dm_target = dm_table->target; dm_target; dm_target = dm_target->next)
+		if (dm_target->start <= sector && sector < (dm_target->start+dm_target->length))
+		    break;
+	    if (!dm_target)
+		die("device-mapper: Sector outside mapped device? (%d: %u/%"PRIu64")",
+		    geo->base_dev, sector, (uint64_t)(dm_table->target ?
+		      (dm_table->target->start+dm_table->target->length) : 0));
+
+	    dev = dm_target->device;
+	    sector = dm_target->offset+(sector-dm_target->start);
+
+	    for(i = 0; i < dm_major_nr; i++)
+		if (MAJOR(dev) == dm_major_list[i])
+		    break;
+	}
+
+	if (dev != geo->dev && dev != geo->base_dev)
+	    die("device-mapper: mapped boot device cannot be on multiple real devices\n");
+#endif
+	sector += geo->start;
+    }
+
  /*   DON'T always use CHS addressing on floppies:     JRC   */
 /*    if ((geo->device & 0x80) && (linear || lba32)) {	*/
     if ((linear || lba32)) {
@@ -1193,9 +1491,8 @@ int geo_comp_addr(GEOMETRY *geo,int offset,SECTOR_ADDR *addr)
 	    if (geo->sectors>0 && geo->heads>0) {
 	    	cyl /= geo->sectors;
 	    	cyl /= geo->heads;
-	    	if (cyl >= BIOS_MAX_CYLS && !nowarn && warnings++ < 8) {
-		    fprintf(errstd,
-			"Warning:  LINEAR may generate cylinder# above 1023 at boot-time.\n");
+	    	if (cyl >= BIOS_MAX_CYLS && linear_warnings++ < 8) {
+		    warn("LINEAR may generate cylinder# above 1023 at boot-time.");
 	    	}
 	    }
             if (sector/(63*255) >= BIOS_MAX_CYLS)
