@@ -163,6 +163,11 @@ void geo_query_dev(GEOMETRY *geo,int device,int all);
    information is then stored in *geo */
 
 
+int is_dm_major(int major);
+/* tell whether the specified major device number is one of the
+   device-mapper major devices */
+   
+
 #if 0
 int geo_devscan(int device);
 /* called to fill in a disktab with arbitrary BIOS codes */
