@@ -1,7 +1,7 @@
 /* lilo.c  -  LILO command-line parameter processing */
 /*
 Copyright 1992-1998 Werner Almesberger.
-Copyright 1999-2005 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
@@ -52,6 +52,159 @@ char *config_file;		/* actual name of the config file */
 int config_read;		/* readable by other than root */
 FILE *errstd;
 
+static void configuration(void)
+{
+#if VERSION_MINOR>=50 && !__MSDOS__
+{
+		printf(
+		"-DBOOT_PARAMS_1=%d -DBOOT_PARAMS_2=%d  PATH_MAX=%d"
+		  "\n"
+		  ,
+		  (int)sizeof(BOOT_PARAMS_1),
+		  (int)sizeof(BOOT_PARAMS_2),
+		  PATH_MAX
+		     );
+}
+{
+#include "bitmap.h"
+	RGB pal[16];
+	RGB2 pal2[16];
+	
+	printf("\nBMFH=%d BMH=%d BMH2=%d RGB=%d(%d) RGB2=%d(%d) LH=%d\n",
+	(int)sizeof(BITMAPFILEHEADER),
+	(int)sizeof(BITMAPHEADER),
+	(int)sizeof(BITMAPHEADER2),
+	(int)sizeof(RGB),
+	(int)sizeof(pal),
+	(int)sizeof(RGB2),
+	(int)sizeof(pal2),
+	(int)sizeof(BITMAPLILOHEADER) );
+	printf("\nBPB=%d BP_DOS=%d\n",
+		(int)sizeof(BIOS_PARAMETER_BLOCK),
+		(int)sizeof(BOOT_PARAMS_DOS)
+		  );
+}
+#ifdef LCF_FIRST6
+	printf("\nSECTOR_ADDR6=%d\n", (int)sizeof(SECTOR_ADDR6));
+#endif
+
+#endif /* VERSION_MINOR>=50 && !__MSDOS__ */
+
+#if !__MSDOS__
+{
+  unsigned int j;
+  if (verbose>=5) {
+    if (crc(First.data,First.size)) {
+	j = crc(First.data,First.size-4);
+	brev(j);
+	printf("1=0x%x\n", j);
+    }
+    if (crc(Second.data,Second.size)) {
+	j = crc(Second.data,Second.size-4);
+	brev(j);
+	printf("2=0x%x\n", j);
+    }
+    if (crc(Third.data,Third.size)) {
+	j = crc(Third.data,Third.size-4);
+	brev(j);
+	printf("3=0x%x\n", j);
+    }
+    if (crc(Bitmap.data,Bitmap.size)) {
+	j = crc(Bitmap.data,Bitmap.size-4);
+	brev(j);
+	printf("B=0x%x\n", j);
+    }
+    if (crc(Chain.data,Chain.size)) {
+	j = crc(Chain.data,Chain.size-4);
+	brev(j);
+	printf("C=0x%x\n", j);
+    }
+    if (crc(Mbr.data,Mbr.size)) {
+	j = crc(Mbr.data,Mbr.size-4);
+	brev(j);
+	printf("M=0x%x\n", j);
+    }
+    if (crc(Mbr2.data,Mbr2.size)) {
+	j = crc(Mbr2.data,Mbr2.size-4);
+	brev(j);
+	printf("N=0x%x\n\n", j);
+    }
+  }
+}
+
+#if BETA_TEST
+{
+    int i, j, k=0;
+extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
+
+
+    printf("\n");
+    for (i=0; i<256; i++) {
+	if ( has_partitions_beta(MKDEV(i,0)) != has_partitions(MKDEV(i,0)) ) {
+	    printf("Major device = %d is not configured.\n", i);
+	    k ++;
+	}
+    }
+    printf("%sk = %d (should be 0)\n\n", k?"ERROR: ":"", k);
+    for (i=0, j=0; i<40960; i+=17) {
+        dev_t dev;
+        int major, minor;
+        for (k=0; k<81920; k+=19) {
+            dev = MKDEV(i,k);
+            major = MAJOR(dev);
+            minor = MINOR(dev);
+            if (major != i  ||  minor != k) {
+                printf("ERROR: (%d,%d) -> (%d,%d)\n",
+                    i, k, major, minor);
+                j++;
+            }
+        }
+    }
+    if (j==0) printf("MKDEV check passed\n");
+}
+#endif
+
+#if VERSION_MINOR>=50
+{
+	dev_t dev; int i;
+	BOOT_VOLID x;
+	BOOT_SECTOR b;
+	dev = 0xFFF0;
+	i = dev;
+	printf("dev_t is %ssigned size=%d  i=%08X\n", i==0xFFF0 ? "un" : "", (int)sizeof(dev_t), i);
+	printf("size of BOOT_VOLID = %d   BOOT_SECTOR = %d\n", (int)sizeof(x), (int)sizeof(b));
+	printf("Size of MENUTABLE = %d\n", (int)sizeof(MENUTABLE));
+}
+#endif
+		printf("\nCFLAGS = " CFLAGS "\n");
+#if defined(LCF_DEVMAPPER) && defined(HAS_LIBDEVMAPPER_H)
+                printf("With");
+#else
+                printf("Without");
+#endif
+                printf(" device-mapper\n");
+		printf("\nglibc version %d.%d\n", __GLIBC__, __GLIBC_MINOR__);
+		printf("Kernel Headers included from  %d.%d.%d\n",
+			LINUX_VERSION_CODE>>16,
+			LINUX_VERSION_CODE>>8 & 255,
+			LINUX_VERSION_CODE & 255);
+		printf("Maximum Major Device = %d\n", MAJOR(~0UL));
+#endif /* !__MSDOS__ */
+		printf("MAX_IMAGES = %d\t\tc=%d, s=%d, i=%d, "
+				"l=%d, ll=%d, f=%d, d=%d, ld=%d\n",
+			MAX_IMAGES, (int)sizeof(char),
+			(int)sizeof(short), (int)sizeof(int),
+			(int)sizeof(long), (int)sizeof(
+#if !__MSDOS__
+			long
+#endif /* !__MSDOS__ */
+ 			     long),
+			(int)sizeof(float), (int)sizeof(double),
+			(int)sizeof(long double)
+			);
+		printf("IMAGE_DESCR = %d   DESCR_SECTORS = %d\n\n",
+			(int)sizeof(IMAGE_DESCR), (int)sizeof(DESCR_SECTORS));
+}
 
 #if !__MSDOS__
 
@@ -116,6 +269,7 @@ static char buffer[SECTOR_SIZE];
     int tsecs;
     int tlinear, tlba32;
     unsigned short flags;
+    time_t Time;
 
 #if !__MSDOS__
     fd = geo_open(&geo,map_file,O_RDONLY);
@@ -159,7 +313,12 @@ static char buffer[SECTOR_SIZE];
 		die("Cannot recognize boot sector.");
 	}
 #endif
+#if 1
+        Time = boot.par_1.map_stamp;
+	printf("Installed:  %s\n", ctime(&Time));
+#else
 	printf("Installed:  %s\n", ctime((time_t*)&boot.par_1.map_stamp));
+#endif
 	printf("Global settings:\n");
 	tsecs = (param2.delay*2197+3999)/4000;
 	printf("  Delay before booting: %d.%d seconds\n",tsecs/10,tsecs % 10);
@@ -170,6 +329,7 @@ static char buffer[SECTOR_SIZE];
 	      tsecs % 10);
 	}
 	printf("  %snattended booting\n", param2.flag2&FLAG2_UNATTENDED ? "U" : "No u");
+	printf("  %sPC/AT keyboard hardware prescence check\n", param2.flag2&FLAG2_NOKBD ? "" : "No ");
 	if (boot.par_1.prompt & FLAG_PROMPT) printf("  Always enter boot prompt\n");
 	else printf("  Enter boot prompt only on demand\n");
 	printf("  Boot-time BIOS data%s saved\n",
@@ -213,10 +373,14 @@ static char buffer[SECTOR_SIZE];
 #if __MSDOS__
 	    printf("%s\n", name
 #else /* !__MSDOS__ */
-	    fprintf(stdout,"%s%-" S(MAX_IMAGE_NAME) "s %s%s",verbose > 0 ? "  " : "",name,
+	    printf("%s%-" S(MAX_IMAGE_NAME) "s %s%s%s",verbose > 0 ? "  " : "",name,
 	      image ? "" : "*",
 #ifdef LCF_VIRTUAL
 	      descrs.d.descr[image].flags & FLAG_VMDEFAULT ? "@" :
+#endif
+	      "",
+#ifdef LCF_NOKEYBOARD
+	      descrs.d.descr[image].flags & FLAG_NOKBDEFAULT ? "&" :
 #endif
 	      ""
 #endif /* !__MSDOS__ */
@@ -252,6 +416,10 @@ static char buffer[SECTOR_SIZE];
 		    printf("    Virtual Boot is disabled\n");
 		if (flags & FLAG_VMWARN)
 		    printf("    Warn on Virtual boot\n");
+#endif		
+#ifdef LCF_NOKEYBOARD
+		if (flags & FLAG_NOKBDISABLE)
+		    printf("    NoKeyboard Boot is disabled\n");
 #endif		
 		if ( !(flags & FLAG_PASSWORD) )
 		    printf("    No password\n");
@@ -485,6 +653,9 @@ fprintf(errstd,"argc=%d, *argv=%s, ch=%c param=%s\n", argc, *argv, ch, param);
 	    case 'g':
 		geometric |= AD_GEOMETRIC;
 		break;
+	    case 'H':
+		force_raid=1;
+		break;
 	    case 'i':
 		cfg_set(cf_options,"install",param,NULL);
 		break;
@@ -617,152 +788,7 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 		break;
 #endif /* !__MSDOS__ */
 	    case 'X':
-#if VERSION_MINOR>=50 && !__MSDOS__
-{
-		printf(
-		"-DBOOT_PARAMS_1=%d -DBOOT_PARAMS_2=%d  PATH_MAX=%d"
-		  "\n"
-		  ,
-		  (int)sizeof(BOOT_PARAMS_1),
-		  (int)sizeof(BOOT_PARAMS_2),
-		  PATH_MAX
-		     );
-}
-{
-#include "bitmap.h"
-	RGB pal[16];
-	RGB2 pal2[16];
-	
-	printf("\nBMFH=%d BMH=%d BMH2=%d RGB=%d(%d) RGB2=%d(%d) LH=%d\n",
-	(int)sizeof(BITMAPFILEHEADER),
-	(int)sizeof(BITMAPHEADER),
-	(int)sizeof(BITMAPHEADER2),
-	(int)sizeof(RGB),
-	(int)sizeof(pal),
-	(int)sizeof(RGB2),
-	(int)sizeof(pal2),
-	(int)sizeof(BITMAPLILOHEADER) );
-	printf("\nBPB=%d BP_DOS=%d\n",
-		(int)sizeof(BIOS_PARAMETER_BLOCK),
-		(int)sizeof(BOOT_PARAMS_DOS)
-		  );
-}
-#ifdef LCF_FIRST6
-	printf("\nSECTOR_ADDR6=%d\n", (int)sizeof(SECTOR_ADDR6));
-#endif
-
-#endif /* VERSION_MINOR>=50 && !__MSDOS__ */
-
-#if !__MSDOS__
-{
-  unsigned int j;
-  if (verbose>=5) {
-    if (crc(First.data,First.size)) {
-	j = crc(First.data,First.size-4);
-	brev(j);
-	printf("1=0x%x\n", j);
-    }
-    if (crc(Second.data,Second.size)) {
-	j = crc(Second.data,Second.size-4);
-	brev(j);
-	printf("2=0x%x\n", j);
-    }
-    if (crc(Third.data,Third.size)) {
-	j = crc(Third.data,Third.size-4);
-	brev(j);
-	printf("3=0x%x\n", j);
-    }
-    if (crc(Bitmap.data,Bitmap.size)) {
-	j = crc(Bitmap.data,Bitmap.size-4);
-	brev(j);
-	printf("B=0x%x\n", j);
-    }
-    if (crc(Chain.data,Chain.size)) {
-	j = crc(Chain.data,Chain.size-4);
-	brev(j);
-	printf("C=0x%x\n", j);
-    }
-    if (crc(Mbr.data,Mbr.size)) {
-	j = crc(Mbr.data,Mbr.size-4);
-	brev(j);
-	printf("M=0x%x\n", j);
-    }
-    if (crc(Mbr2.data,Mbr2.size)) {
-	j = crc(Mbr2.data,Mbr2.size-4);
-	brev(j);
-	printf("N=0x%x\n\n", j);
-    }
-  }
-}
-
-#if BETA_TEST
-{
-    int i, j, k=0;
-extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
-
-
-    printf("\n");
-    for (i=0; i<256; i++) {
-	if ( has_partitions_beta(MKDEV(i,0)) != has_partitions(MKDEV(i,0)) ) {
-	    printf("Major device = %d is not configured.\n", i);
-	    k ++;
-	}
-    }
-    printf("%sk = %d (should be 0)\n\n", k?"ERROR: ":"", k);
-    for (i=0, j=0; i<40960; i+=17) {
-        dev_t dev;
-        int major, minor;
-        for (k=0; k<81920; k+=19) {
-            dev = MKDEV(i,k);
-            major = MAJOR(dev);
-            minor = MINOR(dev);
-            if (major != i  ||  minor != k) {
-                printf("ERROR: (%d,%d) -> (%d,%d)\n",
-                    i, k, major, minor);
-                j++;
-            }
-        }
-    }
-    if (j==0) printf("MKDEV check passed\n");
-}
-#endif
-
-#if VERSION_MINOR>=50
-{
-	dev_t dev; int i;
-	BOOT_VOLID x;
-	BOOT_SECTOR b;
-	dev = 0xFFF0;
-	i = dev;
-	printf("dev_t is %ssigned size=%d  i=%08X\n", i==0xFFF0 ? "un" : "", (int)sizeof(dev_t), i);
-	printf("size of BOOT_VOLID = %d   BOOT_SECTOR = %d\n", (int)sizeof(x), (int)sizeof(b));
-	printf("Size of MENUTABLE = %d\n", (int)sizeof(MENUTABLE));
-}
-#endif
-		printf("\nCFLAGS = " CFLAGS "\n");
-		printf("\nglibc version %d.%d\n", __GLIBC__, __GLIBC_MINOR__);
-		printf("Kernel Headers included from  %d.%d.%d\n",
-			LINUX_VERSION_CODE>>16,
-			LINUX_VERSION_CODE>>8 & 255,
-			LINUX_VERSION_CODE & 255);
-		printf("Maximum Major Device = %d\n", MAJOR(~0UL));
-#endif /* !__MSDOS__ */
-		printf("\nMAX_IMAGES = %d\t\tc=%d, s=%d, i=%d, "
-				"l=%d, ll=%d, f=%d, d=%d, ld=%d\n\n",
-			MAX_IMAGES, (int)sizeof(char),
-			(int)sizeof(short), (int)sizeof(int),
-			(int)sizeof(long), (int)sizeof(
-#if !__MSDOS__
-			long
-#endif /* !__MSDOS__ */
- 			     long),
-			(int)sizeof(float), (int)sizeof(double),
-			(int)sizeof(long double)
-			);
-#if __MSDOS__ || 1
-		printf("\nIMAGE_DESCR = %d   DESCR_SECTORS = %d\n",
-			(int)sizeof(IMAGE_DESCR), (int)sizeof(DESCR_SECTORS));
-#endif /* __MSDOS__ */
+	        configuration();
 		exit(0);
 #if !__MSDOS__
 	    case 'z':
@@ -782,8 +808,8 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     if (new_root && *new_root) {
 	pp_fd = fopen(PARTITIONS, "r");
 	if (chroot(new_root) < 0) die("chroot %s: %s",new_root,strerror(errno));
-	if (chdir("/dev") < 0  &&  nowarn < 1 )
-	    fprintf(errstd, "Warning: root at %s has no /dev directory\n", new_root);
+	if (chdir("/dev") < 0)
+	        warn("root at %s has no /dev directory", new_root);
 	if (chdir("/") < 0) die("chdir /: %s",strerror(errno));
     }
     if (atexit(temp_remove)) die("atexit() failed");
@@ -798,23 +824,27 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 	    return 0;
 	}
 	printf(", Copyright (C) 1992-1998 Werner Almesberger\n"
-	       "Development beyond version 21 Copyright (C) 1999-2005 John Coffman\n"
+	       "Development beyond version 21 Copyright (C) 1999-2006 John Coffman\n"
 	       );
         if (verbose>0) {
-#if VERSION_MINOR>=50 && !__MSDOS__
+#if !__MSDOS__
 #include <sys/utsname.h>
 	    struct utsname buf;
 #endif
             printf("Released %s%s and compiled at %s on %s%s\n",
 		VERSION_DATE, comma ? "," : "", __TIME__, __DATE__, semi);
-#if VERSION_MINOR>=50 && !__MSDOS__
-	    if (uname(&buf)==0) {
-		printf("Running %s kernel %s\n", buf.sysname, buf.release);
+#if !__MSDOS__
+	    if (verbose>=2 && uname(&buf)==0) {
+		printf("Running %s kernel %s on %s\n",
+		        buf.sysname, buf.release, buf.machine);
 	    }
 #endif
 	}
         printf("\n");
-        if (version) return 0;
+        if (version) {
+            if (verbose>=2) configuration();
+            return 0;
+        }
     }
 
     if (verbose > 0) errstd = stdout;
@@ -835,31 +865,29 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     
 #if !__MSDOS__
     temp = cfg_get_flag(cf_options,"nowarn");
-#endif /* !__MSDOS__ */
     if (nowarn < 0) nowarn = 0;
     else nowarn = temp;
+/* All warnings appear if very verbose modes used */
+    if (verbose>=3) nowarn = 0;
+#endif /* !__MSDOS__ */
 
     if (verbose>=6) printf("main: cfg_parse returns %d\n", more);
 
 #if !__MSDOS__
     if (tell_param && !tell_early) probe_tell(tell_param);
 
-    if (!nowarn) {
-	if (fstat(fd,&st) < 0) {
-	    fprintf(errstd,"fstat %s: %s\n",config_file,strerror(errno));
-	    exit(1);
-	}
-	if (S_ISREG(st.st_mode)) {
-	    if (st.st_uid)
-		fprintf(errstd,"Warning: %s should be owned by root\n",
-		  config_file);
-	    else if (st.st_mode & (S_IWGRP | S_IWOTH))
-		    fprintf(errstd,"Warning: %s should be writable only for "
-		      "root\n",config_file);
-	    config_read = !!(st.st_mode & (S_IRGRP | S_IROTH));
-	}
+    if (fstat(fd,&st) < 0)
+	    die("fstat %s: %s", config_file, strerror(errno) );
+
+    if (S_ISREG(st.st_mode)) {
+        if (st.st_uid)
+            warn("%s should be owned by root", config_file);
+        else if (st.st_mode & (S_IWGRP | S_IWOTH))
+	    warn("%s should be writable only for root", config_file);
+        config_read = !!(st.st_mode & (S_IRGRP | S_IROTH));
     }
-    preload_dev_cache();
+
+    if (!cfg_get_flag(cf_options,"nodevcache"))  preload_dev_cache();
     
     if (verbose<0 && cfg_get_strg(cf_options,"verbose"))
 	verbose = to_number(cfg_get_strg(cf_options,"verbose"));
@@ -874,23 +902,23 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     if (geom+linear+lba32 > 1)
 	die("May specify only one of GEOMETRIC, LINEAR or LBA32");
     if (geometric) {
-	if (!nowarn && (geom+linear+lba32 > 0))  
-	    fprintf(errstd,"Ignoring entry '%s'\n", geom ? "geometric" :
-	    	linear ? "linear" : "lba32");
+	if (geom+linear+lba32 > 0)  
+	    warn("Ignoring entry '%s'", geom ? "geometric" :
+	    	      linear ? "linear" : "lba32");
 	geom = linear = lba32 = 0;
 	if (geometric==AD_LBA32) lba32 = 1;
 	else if (geometric==AD_LINEAR) linear = 1;
 	else if (geometric==AD_GEOMETRIC) geom = 1;
     }    
     if (geom+linear+lba32 == 0) {
-	if (!nowarn && verbose) fprintf(errstd,"Warning: LBA32 addressing assumed\n");
+	warn("LBA32 addressing assumed");
 	lba32 = 1;
     }
-    if (linear && !nowarn) fprintf(errstd,
-    	"\nWarning:  LINEAR is deprecated in favor of LBA32:  LINEAR specifies 24-bit\n"
+    if (linear) warn(
+    	"LINEAR is deprecated in favor of LBA32:  LINEAR specifies 24-bit\n"
 	"  disk addresses below the 1024 cylinder limit; LBA32 specifies 32-bit disk\n"
-	"  addresses not subject to cylinder limits on systems with EDD bios extensions;\n"
-	"  use LINEAR only if you are aware of its limitations.\n\n");
+	"  addresses not subject to cylinder limits on systems with EDD-BIOS extensions;\n"
+	"  use LINEAR only if you are aware of its limitations.");
     
     if (identify) identify_image(identify,ident_opt);
 
@@ -912,9 +940,9 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
     }
     if (bios_passes_dl == DL_NOT_SET) 	check_bios();	/* in probe.c */
 
-    if (!nowarn && compact && (linear || lba32) && verbose>=2)
-	fprintf(errstd,"Warning: COMPACT may conflict with %s on some "
-		"systems\n", lba32 ? "LBA32" : "LINEAR");
+    if (compact && (linear || lba32) && verbose>=4)
+	warn("COMPACT may conflict with %s on some "
+		"systems", lba32 ? "LBA32" : "LINEAR");
 
     geo_init(cfg_get_strg(cf_options,"disktab"));
 #endif /* !__MSDOS__ */
@@ -958,6 +986,9 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 
 #ifdef LCF_VIRTUAL
 	check_vmdefault();
+#endif
+#ifdef LCF_NOKEYBOARD
+	check_nokbdefault();
 #endif
 	check_fallback();
 	check_unattended();
@@ -983,6 +1014,12 @@ extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
 	      "altered.\n");
 	}
 	if (verbose>=4) dump_serial_nos();
+	if (warnings) {
+	    if (warnings>1)
+	        printf("%d warnings were ", warnings);
+            else printf("One warning was ");
+            printf("%sed.\n", nowarn ? "suppress" : "issu");
+	}
 #else  /* __MSDOS__ */
 	die("No option switches specified:  -q, -R, or -V");
 #endif /* __MSDOS__ */
