@@ -203,8 +203,8 @@ void part_verify(int dev_nr,int type)
     }
 #if 1
     if (type && pe>0 && pe<=(mbr==STAGE_MBR2?63:PART_MAX)
-    	     && !(part_table[part].boot_ind&0x80) && !nowarn)
-	fprintf(errstd,"Warning: Partition %d on %s is not marked Active.\n",
+    	     && !(part_table[part].boot_ind&0x80) )
+	warn("Partition %d on %s is not marked Active.",
 		pe, dev.name);
 #endif
     i = part_table[part].sys_ind;
@@ -219,16 +219,17 @@ void part_verify(int dev_nr,int type)
 	      i == PART_DOS16_BIG ||
 	      i == PART_FAT32 ||
 	      i == PART_FAT32_LBA ||
-	      i == PART_FAT16_LBA ;
+	      i == PART_FAT16_LBA ||
+	      i == PART_NTFS ||
+	      i == PART_OS2_BOOTMGR ;
 
     if (type && !Linux) {
-	fflush(stdout);
-	if (!nowarn) fprintf(errstd,"Warning: partition type 0x%02X"" on device 0x%04X is a dangerous place for\n"
-							"    a boot sector.",
-			part_table[part].sys_ind, dev_nr);
-	if (dos && !nowarn) fprintf(errstd,"  A DOS/Windows/OS2 system may be rendered unbootable."
-		"\n  The backup copy of this boot sector should be retained.");
-	if (!nowarn) fprintf(errstd,"\n");
+	warn("partition type 0x%02X"" on device 0x%04X is a dangerous place for\n"
+             "    a boot sector.%s",
+			part_table[part].sys_ind, dev_nr,
+	dos ? "  A DOS/Windows/OS2 system may be rendered unbootable."
+		"\n  The backup copy of this boot sector should be retained."
+		: "" );
 #if 0
 	if (!dos && !cfg_get_flag(cf_options,"ignore-table"))
 	    die("You may proceed by using either '-P ignore' or 'ignore-table'");
@@ -241,18 +242,18 @@ void part_verify(int dev_nr,int type)
       cyl*geo.heads)*geo.sectors;
     if (pe <= PART_MAX &&
 	    (lin_3d > part_table[part].start_sect || (lin_3d <
-	    part_table[part].start_sect && cyl != BIOS_MAX_CYLS-1)) &&
-	    !nowarn) {
-	fflush(stdout);
-	fprintf(errstd,"Device 0x%04X: Inconsistent partition table, %d%s entry\n",
+	    part_table[part].start_sect && cyl != BIOS_MAX_CYLS-1)) ) {
+	warn("Device 0x%04X: Inconsistent partition table, %d%s entry",
 	  dev_nr & ~mask,part+1,!part ? "st" : part == 1 ? "nd" : part ==
 	  2 ? "rd" : "th");
+        if (!nowarn)
 	fprintf(errstd,"  CHS address in PT:  %d:%d:%d  -->  LBA (%d)\n",
 		cyl,
 		part_table[part].head,
 		part_table[part].sector & 63,
 		lin_3d);
 	cyl = part_table[part].start_sect/geo.sectors/geo.heads;
+        if (!nowarn)
 	fprintf(errstd,"  LBA address in PT:  %d  -->  CHS (%d:%d:%d)\n",
 		part_table[part].start_sect,
 		cyl,
@@ -265,9 +266,9 @@ void part_verify(int dev_nr,int type)
 	if (!cfg_get_flag(cf_options,"fix-table") && !cfg_get_flag(cf_options,
 	  "ignore-table")) die("Either FIX-TABLE or IGNORE-TABLE must be specified\n"
 			"If not sure, first try IGNORE-TABLE (-P ignore)");
-	if (test || cfg_get_flag(cf_options,"ignore-table"))
-	    fprintf(errstd,"The partition table is *NOT* being adjusted.\n");
-	else {
+	if (test || cfg_get_flag(cf_options,"ignore-table")) {
+	    warn("The partition table is *NOT* being adjusted.");
+	} else {
 	    sprintf(backup_file,BACKUP_DIR "/part.%04X",dev_nr & ~mask);
 	    if ((bck_file = creat(backup_file,0644)) < 0)
 		die("creat %s: %s",backup_file,strerror(errno));
@@ -465,9 +466,8 @@ void do_cr_auto(void)
     	CHANGE_RULE *cr;
 	if ((cr=may_change(part_table[i].sys_ind))) {
 	    j = i*PARTITION_ENTRY + PART_TYPE_ENT_OFF;
-	    if (autoauto && !nowarn) {
-	    	fflush(stdout);
-	        fprintf(errstd, "Warning: CHANGE AUTOMATIC assumed after \"other=%s\"\n", other);
+	    if (autoauto) {
+	        warn("CHANGE AUTOMATIC assumed after \"other=%s\"", other);
 	        autoauto = 0;  /* suppress further warnings */
 	    }    
 	    if (i == partition-1)
@@ -477,9 +477,8 @@ void do_cr_auto(void)
 	}
     }
 #else
-    if (!nowarn) fprintf(errstd,
-       "Warning:  This LILO is compiled without REWRITE_TABLE;\n"
-       "   unable to generate CHANGE/AUTOMATIC change-rules\n");
+    warn("This LILO is compiled without REWRITE_TABLE;\n"
+       "   unable to generate CHANGE/AUTOMATIC change-rules");
 #endif
 }
 
