@@ -1,6 +1,6 @@
 /* probe.c -- BIOS probes */
 /*
-Copyright 1999-2005 John Coffman.
+Copyright 1999-2006 John Coffman.
 All rights reserved.
 
 Licensed under the terms contained in the file 'COPYING' in the 
@@ -8,7 +8,7 @@ source directory.
 
 */
 
-/*#define DEBUG*/
+/*#define DEBUG_PROBE*/
 #define BITMAP 0	/* change to 1 when do_bitmap is filled in */
 #define VOLID  1	/* change to 1 when do_volid is filled in */
 
@@ -22,6 +22,7 @@ source directory.
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include "config.h"
 #include "lilo.h"
 #include "common.h"
 #include "device.h"
@@ -130,7 +131,7 @@ struct Probes {
 { "EBDA",  do_ebda,  NULL,  "Extended BIOS Data Area information" },
 { "geom=", do_geom,  "<bios>", "Geometry CHS data for BIOS code 0x80, etc." },
 { "geom" , do_geom_all, NULL, "Geometry for all BIOS drives" },
-{ "table=", do_table, dev, "Partition table information for /dev/hda, etc."},
+{ "table=", do_table, dev, "Partition table information for " DEV_DISK_DIR "/hda, etc."},
 { "video", do_video, NULL,  "Graphic mode information" },
 #if VOLID
 { "vol-ID", do_volid, NULL, "Volume ID check for uniqueness"},
@@ -176,7 +177,7 @@ int fetch(void)
     
     if (buf_valid>=0) return buf_valid;
     
-    if ((fd=open("/dev/mem", O_RDONLY)) < 0) return buf_valid=1;
+    if ((fd=open(DEV_DIR "/mem", O_RDONLY)) < 0) return buf_valid=1;
     at = lseek(fd, seek, SEEK_SET);
     if (at != seek) return buf_valid=1;
     get = sizeof(buf.b);
@@ -231,7 +232,7 @@ int purge(void)
     i = buf.s5.length;
     *(int*)buf.s5.checksum = crc32((unsigned char*)&buf.s5 + 4, i-4, CRC_POLY1);
     
-    if ((fd=open("/dev/mem", O_WRONLY)) < 0) pdie("purge: can't open /dev/mem");
+    if ((fd=open(DEV_DIR "/mem", O_WRONLY)) < 0) pdie("purge: can't open " DEV_DIR "/mem");
     if (lseek(fd, seek, SEEK_SET) != seek) pdie("purge: ");
     i = sizeof(buf.b);
     if (write(fd, &buf.b, i) != i) pdie("purge: ");
@@ -259,7 +260,7 @@ static int notice(int needed)
 		"Booting your system with LILO version %s or later would provide the re-\n"
 		"quested information as part of the BIOS data check.  Please install a more\n"
 		"recent version of LILO on your hard disk, or create a bootable rescue floppy\n"
-		"with the 'mkrescue' command.\n\n",
+		"or rescue CD with the 'mkrescue' command.\n\n",
 		
 		needed==4 ? "22.0" :
 		needed==5 ? "22.5.1" :
@@ -283,7 +284,7 @@ static void do_help(void)
     			pr->str ? pr->str : "        ",
     			pr->help);
     }
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
     printf("    In some cases, the '-v' flag will produce more output.\n");
     printf("sizeof(video_t) = %d  sizeof(floppy_t) = %d  sizeof(hard_t) = %d\n"
            "sizeof(edd_t) = %d  sizeof(buf.s) = %d\n",
@@ -297,6 +298,7 @@ static void do_help(void)
 /* diagnostic output */
 static void show_geom(char *who, int cyl, int head, int sect)
 {
+   if (nowarn) return;
    fprintf(errstd, "    %s: %d cylinders, %d heads, %d sectors\n", who, cyl, head, sect);
 }
 
@@ -309,6 +311,8 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
     int i;
     struct partition *pt_base;
     unsigned int total;
+    int sec_err = 0;
+    int hd_err = 0;
 
 #if 0
     if((i=fetch())) {
@@ -318,7 +322,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 #else
     if (notice(4)) exit(0);
 #endif
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	printf("get_geom: drive = 0x%02X\n", drive);
 	fflush(stdout);
 #endif
@@ -333,7 +337,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 	if (drive > buf.s5.disk) return 1;
 	if (!hdp[drive-0x80]) {
 	    i = get_geom(drive-1, geom);
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 		printf("get_geom recursive return = %d  AH=0x%02X\n", i, i-1);
 		fflush(stdout);
 #endif
@@ -341,7 +345,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 	}
 	hd = hdp[drive-0x80];
     } else return 1;
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	printf("get_geom:  hd = %08X\n", (int)hd);
 	fflush(stdout);
 #endif
@@ -355,7 +359,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
     /* regs.eax = 0x1500;           check drive type */
     /* regs.edx = drive;			*/
 
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	printf("get_geom: int13, fn=15\n");
 	fflush(stdout);
 #endif
@@ -369,7 +373,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
    /* regs.eax = 0x0800;		*/
    /* regs.edx = drive;			*/
    
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	printf("get_geom: int13, fn=08\n");
 	fflush(stdout);
 #endif
@@ -377,8 +381,9 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
    if (hd->fn08.flags&1 || hd->fn08.ah || hd->fn08.cx==0)
      return 1 + hd->fn08.ah;
    
-   i =
-   geom->n_sect = hd->fn08.cx & 0x3F;
+   if (!(i = hd->fn08.cx & 0x3F))  i = 64;	/* BIOS bug if 0 */
+   geom->n_sect = i;
+   
    i *=
    geom->n_head = ((hd->fn08.dx>>8)&0xFF)+1;
    i *=
@@ -387,6 +392,9 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
    geom->n_disks = hd->fn08.dx & 0xFF;
    geom->pt = NULL;
 
+   hd_err = (geom->n_head > 255);
+   sec_err = (geom->n_sect > 63);
+   
    if (drive < 4)  return 0;
    
    pt_base = NULL;
@@ -406,14 +414,14 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 #endif
    }
 
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
    printf("get_geom:  PT->%08X  S/N=%08X\n", (int)geom->pt, geom->serial_no);
 #endif
       
    /* regs.eax = 0x4100;      check EDD extensions present */
    /* regs.edx = drive;				*/
    /* regs.ebx = 0x55AA;			*/
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	printf("get_geom: int13, fn=41\n");
 	fflush(stdout);
 #endif
@@ -426,7 +434,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
       edd_t *dp;
 
       dp = (edd_t*)hdp[drive-0x80 + 1];
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	printf("get_geom:  EDD  dp = %08X\n", (int)dp);
 	fflush(stdout);
 #endif
@@ -436,7 +444,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
       /* regs.eax = 0x4800;		*/
       /* regs.edx = drive;		*/
       
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	printf("get_geom: int13, fn=48\n");
 	fflush(stdout);
 #endif
@@ -446,9 +454,9 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 	      if ((dp->info) & EDD_PARAM_GEOM_VALID) {
 	         if ((geom->n_sect != dp->sectors || geom->n_head != dp->heads) &&
 				((verbose>0 && !lba32) || verbose>=4) && 
-				!nowarn && !(warned[drive-0x80]&1) ) {
-	                fprintf(errstd,"Warning: Int 0x13 function 8 and function 0x48 return different\n"
-                	               "head/sector geometries for BIOS drive 0x%02X\n", drive);
+				!(warned[drive-0x80]&1) ) {
+	                warn("Int 0x13 function 8 and function 0x48 return different\n"
+                	               "head/sector geometries for BIOS drive 0x%02X", drive);
         	        show_geom("fn 08", geom->n_cyl, geom->n_head, geom->n_sect);
 	                show_geom("fn 48", dp->cylinders, dp->heads, dp->sectors);
 			warned[drive-0x80] |= 1;
@@ -469,7 +477,20 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 	      if (dp->total_sectors > geom->n_total_blocks)
 			geom->n_total_blocks = dp->total_sectors;
 	}
-   }
+    }
+    if (!(warned[drive-0x80]&4)) {
+       if (hd_err) warn("LILO is compensating for a BIOS bug: (drive 0x%02X) heads > 255",
+                           drive);
+       if (sec_err) {
+           warn("LILO will try to compensate for a BIOS bug: (drive 0x%02X) sectors > 63",
+                           drive);
+           if ((geom->EDD_flags & EDD_PACKET) && !lba32)
+               die("LBA32 addressing should be used, not %s", linear ? "LINEAR" : "GEOMETRIC");
+           if (!(geom->EDD_flags & EDD_PACKET)   &&   !(lba32 | linear) )
+               warn("Drive 0x%02X may not be usable at boot-time.", drive);
+       }
+       warned[drive-0x80] |= 4;
+    }
    
    return 0;
 }
@@ -507,7 +528,7 @@ static void do_ebda(void)
     m <<= 10;
     m -= 0x200;
     n = (select_loader()->size + SECTOR_SIZE - 1) / SECTOR_SIZE;
-    n = m - (n+4+MAX_DESCR_SECS)*SECTOR_SIZE;
+    n = m - (n+4+MAX_DESCR_SECS+(COMMAND_LINE_SIZE>256))*SECTOR_SIZE;
     init = (n - (MAX_SETUPSECS+1)*SECTOR_SIZE)>>4;
     if (init > DEF_INITSEG) init = DEF_INITSEG;
     printf("\n");
@@ -549,7 +570,7 @@ static void print_geom(int dr, struct disk_geom geom)
     char *sp = sz;
     unsigned int n, m=0;
 
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
     if (!dr) geom.n_total_blocks = 4000000000UL;	/* 2.09Tb */
 #endif
 
@@ -889,7 +910,7 @@ int bios_max_devs(void)
     return BIOS_MAX_DEVS;    
 }
 
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 static void dump_pt(unsigned char *pt)
 {
     int i, j;
@@ -924,7 +945,7 @@ int bios_device(GEOMETRY *geo, int device)
     if (fetch()) return -1;
     
     if (verbose>=5) printf("bios_dev:  device %04X\n", device);
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 	fflush(stdout);
 #endif    
     if (!has_partitions(device)) return -1;
@@ -962,7 +983,7 @@ int bios_device(GEOMETRY *geo, int device)
 	die("bios_device: read partition table");
     dev_close(&dev);
 
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
     if (verbose>=5) {
         printf("serial number = %08X\n", serial);
         dump_pt(part);
@@ -980,7 +1001,7 @@ int bios_device(GEOMETRY *geo, int device)
 	if (verbose>=5) {
 		printf("bios_dev: (0x%02X)  vol-ID=%08X  *PT=%08lX\n",
 			bios, bdata.serial_no, (long)bdata.pt);
-#ifdef DEBUG
+#ifdef DEBUG_PROBE
 		dump_pt((void*)bdata.pt);
 #endif
 	}
@@ -1008,8 +1029,8 @@ int bios_device(GEOMETRY *geo, int device)
 	       (geo->heads && geo->heads!=bdata.n_head)  )  {
 	    unsigned int nblocks = geo->cylinders * geo->heads * geo->sectors;
 
-	    if (verbose>0 && !nowarn && !(warned[bios1-0x80]&2) ) {
-		fprintf(errstd,"Warning: Kernel & BIOS return differing head/sector geometries for device 0x%02X\n", bios1);
+	    if (!(lba32 | linear) && !(warned[bios1-0x80]&2) ) {
+		warn("Kernel & BIOS return differing head/sector geometries for device 0x%02X", bios1);
 		show_geom("Kernel", geo->cylinders, geo->heads, geo->sectors);
 		show_geom("  BIOS", bdata.n_cyl, bdata.n_head, bdata.n_sect);
 		warned[bios1-0x80] |= 2;
