@@ -144,6 +144,9 @@ int raid_setup(void)
 
 	if (ioctl(md_fd,RAID_VERSION,&md_version_info) < 0)
 	    die("Unable to get RAID version on %s", boot);
+        if (verbose >= 4) printf("RAID_VERSION = %d.%d\n",
+                (int)md_version_info.major,
+                (int)md_version_info.minor);
 	if (md_version_info.major > 0)
 	    die("Raid major versions > 0 are not supported");
 	if (md_version_info.minor < 90)
@@ -151,9 +154,18 @@ int raid_setup(void)
 	
 	if (ioctl(md_fd,GET_ARRAY_INFO,&md_array_info) < 0)
 	    die("Unable to get RAID info on %s",boot);
+        if (verbose >= 4) printf("GET_ARRAY_INFO version = %d.%d\n",
+                (int)md_array_info.major_version,
+                (int)md_array_info.minor_version);
 	if ((md_array_info.major_version != md_version_info.major) &&
-		(md_array_info.minor_version != md_version_info.minor))
-	    die("Inconsistent Raid version information on %s", boot);
+		(md_array_info.minor_version != md_version_info.minor)) {
+	    die("Inconsistent Raid version information on %s   (RV=%d.%d GAI=%d.%d)",
+	        boot,
+                (int)md_version_info.major,
+                (int)md_version_info.minor,
+                (int)md_array_info.major_version,
+                (int)md_array_info.minor_version);
+	    }
 	if (md_array_info.level != 1)
 	    die("Only RAID1 devices are supported as boot devices");
 	if (!linear && !lba32) {
